from curses.ascii import isdigit
from decimal import localcontext
import inspect
import os
import os.path as osp
from pathlib import Path
import re
import tempfile
from threading import local
import warnings
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from aligo import Aligo
from pathlib import Path
from typing import Iterable, Iterator, Optional, Tuple, Union
from antgo.framework.helper.utils.path import is_filepath
from antgo.ant import environment
import shutil
import time


class BaseStorageBackend(metaclass=ABCMeta):
    """Abstract class of storage backends.

    All backends need to implement two apis: ``get()`` and ``get_text()``.
    ``get()`` reads the file as a byte stream and ``get_text()`` reads the file
    as texts.
    """

    # a flag to indicate whether the backend can create a symlink for a file
    _allow_symlink = False

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def allow_symlink(self):
        return self._allow_symlink

    @abstractmethod
    def download(self, remote_path, local_path=None):
        pass

    @abstractmethod
    def upload(self, remote_path, local_path, is_exist=False):
        pass

    @abstractmethod
    def ls(self, remote_folder):
        pass

    @abstractmethod
    def mkdir(self, remote_path, p=False):
        pass


class HardDiskBackend(BaseStorageBackend):
    """Raw hard disks storage backend."""

    _allow_symlink = True

    def get(self, filepath: Union[str, Path]) -> bytes:
        """Read data from a given ``filepath`` with 'rb' mode.

        Args:
            filepath (str or Path): Path to read data.

        Returns:
            bytes: Expected bytes object.
        """
        with open(filepath, 'rb') as f:
            value_buf = f.read()
        return value_buf

    def get_text(self,
                 filepath: Union[str, Path],
                 encoding: str = 'utf-8') -> str:
        """Read data from a given ``filepath`` with 'r' mode.

        Args:
            filepath (str or Path): Path to read data.
            encoding (str): The encoding format used to open the ``filepath``.
                Default: 'utf-8'.

        Returns:
            str: Expected text reading from ``filepath``.
        """
        with open(filepath, 'r', encoding=encoding) as f:
            value_buf = f.read()
        return value_buf

    def put(self, obj: bytes, filepath: Union[str, Path]) -> None:
        """Write data to a given ``filepath`` with 'wb' mode.

        Note:
            ``put`` will create a directory if the directory of ``filepath``
            does not exist.

        Args:
            obj (bytes): Data to be written.
            filepath (str or Path): Path to write data.
        """
        if not os.path.exists(osp.dirname(filepath)):
            os.makedirs(osp.dirname(filepath))
        with open(filepath, 'wb') as f:
            f.write(obj)

    def put_text(self,
                 obj: str,
                 filepath: Union[str, Path],
                 encoding: str = 'utf-8') -> None:
        """Write data to a given ``filepath`` with 'w' mode.

        Note:
            ``put_text`` will create a directory if the directory of
            ``filepath`` does not exist.

        Args:
            obj (str): Data to be written.
            filepath (str or Path): Path to write data.
            encoding (str): The encoding format used to open the ``filepath``.
                Default: 'utf-8'.
        """
        if not os.path.exists(osp.dirname(filepath)):
            os.makedirs(osp.dirname(filepath))        
        with open(filepath, 'w', encoding=encoding) as f:
            f.write(obj)

    def remove(self, filepath: Union[str, Path]) -> None:
        """Remove a file.

        Args:
            filepath (str or Path): Path to be removed.
        """
        os.remove(filepath)

    def exists(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path exists.

        Args:
            filepath (str or Path): Path to be checked whether exists.

        Returns:
            bool: Return ``True`` if ``filepath`` exists, ``False`` otherwise.
        """
        return osp.exists(filepath)

    def isdir(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path is a directory.

        Args:
            filepath (str or Path): Path to be checked whether it is a
                directory.

        Returns:
            bool: Return ``True`` if ``filepath`` points to a directory,
            ``False`` otherwise.
        """
        return osp.isdir(filepath)

    def isfile(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path is a file.

        Args:
            filepath (str or Path): Path to be checked whether it is a file.

        Returns:
            bool: Return ``True`` if ``filepath`` points to a file, ``False``
            otherwise.
        """
        return osp.isfile(filepath)

    def join_path(self, filepath: Union[str, Path],
                  *filepaths: Union[str, Path]) -> str:
        """Concatenate all file paths.

        Join one or more filepath components intelligently. The return value
        is the concatenation of filepath and any members of *filepaths.

        Args:
            filepath (str or Path): Path to be concatenated.

        Returns:
            str: The result of concatenation.
        """
        return osp.join(filepath, *filepaths)

    def list_dir_or_file(self,
                         dir_path: Union[str, Path],
                         list_dir: bool = True,
                         list_file: bool = True,
                         suffix: Optional[Union[str, Tuple[str]]] = None,
                         recursive: bool = False) -> Iterator[str]:
        """Scan a directory to find the interested directories or files in
        arbitrary order.

        Note:
            :meth:`list_dir_or_file` returns the path relative to ``dir_path``.

        Args:
            dir_path (str | Path): Path of the directory.
            list_dir (bool): List the directories. Default: True.
            list_file (bool): List the path of files. Default: True.
            suffix (str or tuple[str], optional):  File suffix
                that we are interested in. Default: None.
            recursive (bool): If set to True, recursively scan the
                directory. Default: False.

        Yields:
            Iterable[str]: A relative path to ``dir_path``.
        """
        if list_dir and suffix is not None:
            raise TypeError('`suffix` should be None when `list_dir` is True')

        if (suffix is not None) and not isinstance(suffix, (str, tuple)):
            raise TypeError('`suffix` must be a string or tuple of strings')

        root = dir_path

        def _list_dir_or_file(dir_path, list_dir, list_file, suffix,
                              recursive):
            for entry in os.scandir(dir_path):
                if not entry.name.startswith('.') and entry.is_file():
                    rel_path = osp.relpath(entry.path, root)
                    if (suffix is None
                            or rel_path.endswith(suffix)) and list_file:
                        yield rel_path
                elif osp.isdir(entry.path):
                    if list_dir:
                        rel_dir = osp.relpath(entry.path, root)
                        yield rel_dir
                    if recursive:
                        yield from _list_dir_or_file(entry.path, list_dir,
                                                     list_file, suffix,
                                                     recursive)

        return _list_dir_or_file(dir_path, list_dir, list_file, suffix,
                                 recursive)

    def ls(self, remote_folder):
        return os.listdir(remote_folder)

    def mkdir(self, remote_path, p=False):
        os.makedirs(remote_path)
        return True

    def download(self, remote_path, local_path):
        # donwload from remote_path to local_path
        if HardDiskBackend._allow_symlink:
            os.system(f'ln -s {remote_path} {local_path}')
        else:
            shutil.copy(remote_path, local_path)

        return True

    def upload(self, remote_path, local_path, is_exist=False):
        # upload from local_path to remote_path
        if is_exist:
            # 如果存在，则取消copy
            if os.path.exists(remote_path):
                return

        shutil.copy(local_path, remote_path)
        return True


class HDFSBackend(BaseStorageBackend):
    _allow_symlink = False

    def get(self, filepath) -> bytes:
        """Read data from a given ``filepath`` with 'rb' mode.

        Args:
            filepath str: Path to read data.

        Returns:
            bytes: Expected bytes object.
        """
        if not os.path.exists('./temp'):
            os.makedirs('./temp')

        environment.hdfs_client.get(filepath, './temp')
        filename = filepath.split('/')[-1]

        with open(os.path.join(f'./temp/{filename}'), 'rb') as f:
            value_buf = f.read()
        return value_buf

    def get_text(self,
                 filepath: str,
                 encoding: str = 'utf-8') -> str:
        """Read data from a given ``filepath`` with 'r' mode.

        Args:
            filepath str: Path to read data.
            encoding (str): The encoding format used to open the ``filepath``.
                Default: 'utf-8'.

        Returns:
            str: Expected text reading from ``filepath``.
        """
        if not os.path.exists('./temp'):
            os.makedirs('./temp')

        environment.hdfs_client.get(filepath, './temp')
        filename = filepath.split('/')[-1]

        with open(os.path.join(f'./temp/{filename}'), 'r', encoding=encoding) as f:
            value_buf = f.read()
        return value_buf

    def put(self, obj: bytes, filepath: str) -> None:
        """Write data to a given ``filepath`` with 'wb' mode.

        Note:
            ``put`` will create a directory if the directory of ``filepath``
            does not exist.

        Args:
            obj (bytes): Data to be written.
            filepath (str or Path): Path to write data.
        """
        if not os.path.exists('./temp'):
            os.makedirs('./temp')
        filename = filepath.split('/')[-1]

        with open(os.path.join('./temp', filename), 'wb') as f:
            f.write(obj)

        p = filepath.find(filename)
        parent_path = filepath[:p]
        if parent_path != '':
            environment.hdfs_client.mkdir(parent_path, True)
        environment.hdfs_client.put(filepath, os.path.join('./temp', filename))

    def put_text(self,
                 obj: str,
                 filepath: str,
                 encoding: str = 'utf-8') -> None:
        """Write data to a given ``filepath`` with 'w' mode.

        Note:
            ``put_text`` will create a directory if the directory of
            ``filepath`` does not exist.

        Args:
            obj (str): Data to be written.
            filepath (str or Path): Path to write data.
            encoding (str): The encoding format used to open the ``filepath``.
                Default: 'utf-8'.
        """
        if not os.path.exists('./temp'):
            os.makedirs('./temp')
        filename = filepath.split('/')[-1]

        with open(os.path.join('./temp', filename), 'w', encoding=encoding) as f:
            f.write(obj)

        p = filepath.find(filename)
        parent_path = filepath[:p]
        if parent_path != '':
            environment.hdfs_client.mkdir(parent_path, True)
        environment.hdfs_client.put(filepath, os.path.join('./temp', filename))

    def remove(self, filepath: str) -> None:
        """Remove a file.

        Args:
            filepath (str or Path): Path to be removed.
        """
        environment.hdfs_client.rm(filepath)

    def exists(self, filepath: str) -> bool:
        """Check whether a file path exists.

        Args:
            filepath (str or Path): Path to be checked whether exists.

        Returns:
            bool: Return ``True`` if ``filepath`` exists, ``False`` otherwise.
        """
        return environment.hdfs_client.exists(filepath)

    def isdir(self, filepath: str) -> bool:
        """
        仅依靠路径名称进行判断，存在不准确
        """
        if filepath.endswith('/'):
            return True
        
        if '.' in filepath.split('/')[-1]:
            return False
        else:
            return True

    def isfile(self, filepath: str) -> bool:
        """
        仅依靠路径名称进行判断，存在不准确
        """
        return not self.isdir(filepath)

    def join_path(self, filepath: str, *filepaths: str) -> str:
        """Concatenate all file paths.

        Join one or more filepath components intelligently. The return value
        is the concatenation of filepath and any members of *filepaths.

        Args:
            filepath (str or Path): Path to be concatenated.

        Returns:
            str: The result of concatenation.
        """
        return osp.join(filepath, *filepaths)

    def list_dir_or_file(self,
                         dir_path: Union[str, Path],
                         list_dir: bool = True,
                         list_file: bool = True,
                         suffix: Optional[Union[str, Tuple[str]]] = None,
                         recursive: bool = False) -> Iterator[str]:
        """Scan a directory to find the interested directories or files in
        arbitrary order.

        Note:
            :meth:`list_dir_or_file` returns the path relative to ``dir_path``.

        Args:
            dir_path (str | Path): Path of the directory.
            list_dir (bool): List the directories. Default: True.
            list_file (bool): List the path of files. Default: True.
            suffix (str or tuple[str], optional):  File suffix
                that we are interested in. Default: None.
            recursive (bool): If set to True, recursively scan the
                directory. Default: False.

        Yields:
            Iterable[str]: A relative path to ``dir_path``.
        """
        raise NotImplementedError

    def ls(self, folder):
        return environment.hdfs_client.ls(folder)

    def mkdir(self, remote_path, p=False):
        environment.hdfs_client.mkdir(remote_path, p)
        return True

    def download(self, remote_path, local_path):
        # donwload from remote_path to local_path
        status = environment.hdfs_client.get(remote_path, local_path)
        return status

    def upload(self, remote_path, local_path, is_exist=False):
        # upload from local_path to remote_path
        status = environment.hdfs_client.put(remote_path, local_path, is_exist)
        return status


class AliBackend(BaseStorageBackend):
    def __init__(self) -> None:
        super().__init__()
        local_config_file = 'aligo.json'
        if os.path.exists(local_config_file):
            if not os.path.exists(Path.home().joinpath('.aligo')):
                os.makedirs(Path.home().joinpath('.aligo'))
                
            shutil.copy(local_config_file, Path.home().joinpath('.aligo'))

        self.ali = Aligo()  # 第一次使用，会弹出二维码，供扫描登录
        if not os.path.exists(local_config_file):
            shutil.copy(os.path.join(Path.home().joinpath('.aligo'), 'aligo.json'),'./')
        self.prefix = 'ali://'
        
    def get(self, filepath) -> bytes:
        if not os.path.exists('./temp'):
            os.makedirs('./temp')

        self.download(filepath, './temp')
        filename = filepath.split('/')[-1]

        with open(os.path.join(f'./temp/{filename}'), 'rb') as f:
            value_buf = f.read()
        return value_buf        
    
    def get_text(self,
                 filepath: str,
                 encoding: str = 'utf-8') -> str:
        if not os.path.exists('./temp'):
            os.makedirs('./temp')

        self.download(filepath, './temp')
        filename = filepath.split('/')[-1]

        with open(os.path.join(f'./temp/{filename}'), 'r', encoding=encoding) as f:
            value_buf = f.read()
        return value_buf        

    def put(self, obj: bytes, filepath: str) -> None:
        if not os.path.exists('./temp'):
            os.makedirs('./temp')
        filename = filepath.split('/')[-1]

        with open(os.path.join('./temp', filename), 'wb') as f:
            f.write(obj)

        p = filepath.find(filename)
        parent_path = filepath[:p]
        self.upload(parent_path, os.path.join('./temp', filename))

    def put_text(self,
                 obj: str,
                 filepath: str,
                 encoding: str = 'utf-8') -> None:
        if not os.path.exists('./temp'):
            os.makedirs('./temp')
        filename = filepath.split('/')[-1]

        with open(os.path.join('./temp', filename), 'w', encoding=encoding) as f:
            f.write(obj)

        p = filepath.find(filename)
        parent_path = filepath[:p]
        self.upload(parent_path, os.path.join('./temp', filename))
    
    def remove(self, filepath: str) -> None:
        pass

    def exists(self, filepath: str) -> bool:
        filepath = filepath.replace(self.prefix, '')
        file = self.ali.get_file_by_path(filepath)
        if file is None:
            file = self.ali.get_folder_by_path(filepath)
            if file is None:
                return False

        return True

    def isdir(self, filepath: str) -> bool:
        file = self.ali.get_folder_by_path(filepath)
        if file is None:
            return False

        return True
    
    def isfile(self, filepath: str) -> bool:
        file = self.ali.get_file_by_path(filepath)
        if file is None:
            return False
        
        return True
    
    def join_path(self, filepath: str, *filepaths: str) -> str:
        """Concatenate all file paths.

        Join one or more filepath components intelligently. The return value
        is the concatenation of filepath and any members of *filepaths.

        Args:
            filepath (str or Path): Path to be concatenated.

        Returns:
            str: The result of concatenation.
        """
        return osp.join(filepath, *filepaths)
    
    def ls(self, remote_folder):
        remote_folder = remote_folder.replace(self.prefix, '')
        file_handler = self.ali.get_folder_by_path(remote_folder)
        if file_handler is None:
            return []

        ll = self.ali.get_file_list(file_handler.file_id)
        ll = [f'{self.prefix}{os.path.join(remote_folder, l.name)}' for l in ll]
        return ll

    def mkdir(self, remote_path, p=False):
        # remote prefix
        remote_path = remote_path.replace(self.prefix, '')
        remote_path = os.path.normpath(remote_path)
        # 迭代创建目录
        levels = remote_path.split('/')[1:]
        level_num = len(levels)
        find_file = None
        find_i = 0
        for i in range(level_num,0,-1):
            check_path = '/'+'/'.join(levels[:i])            
            find_file = self.ali.get_folder_by_path(check_path)
            if find_file:
                break
            find_i = i

        if find_i == 0:
            # 已经存在，不进行重新创建
            return find_file.file_id

        sub_folder = '/'.join(levels[find_i-1:])
        ss = self.ali.create_folder(sub_folder,find_file.file_id)
        return ss.file_id

    def download(self, remote_path, local_path):
        # user = ali.get_user
        remote_path = remote_path.replace(self.prefix, '')
        download_as_multi_files = False
        if remote_path.endswith('*'):
            download_as_multi_files = True
            file = self.ali.get_folder_by_path(os.path.dirname(remote_path))
        else:
            file = self.ali.get_file_by_path(remote_path)
            if file is None:
                file = self.ali.get_folder_by_path(remote_path)

        if file is None:
            return False

        if file.type == 'file':
            self.ali.download_file(file=file, local_folder=local_path)
        else:
            if not download_as_multi_files:
                self.ali.download_folder(file.file_id, local_folder=local_path)
            else:
                ll = self.ali.get_file_list(file.file_id)
                prefix = remote_path.split('/')[-1][:-1]
                filter_ll = []
                if prefix == '':
                    filter_ll = ll
                else:
                    for l in ll:
                        if l.name.startswith(prefix):
                            filter_ll.append(l)
                self.ali.download_files(filter_ll, local_folder=local_path)

        return True
    
    def upload(self, remote_path, local_path, is_exist=False):
        # 检查远程目录是否存在
        file_id = self.mkdir(remote_path, True)
        if file_id is None:
            # error
            return False

        if os.path.isdir(local_path):
            # 目录
            self.ali.upload_folder(local_path, file_id)
        else:
            # 文件
            self.ali.upload_file(local_path, file_id)
        
        return True



class FileClient:
    """A general file client to access files in different backends.

    The client loads a file or text in a specified backend from its path
    and returns it as a binary or text file. There are two ways to choose a
    backend, the name of backend and the prefix of path. Although both of them
    can be used to choose a storage backend, ``backend`` has a higher priority
    that is if they are all set, the storage backend will be chosen by the
    backend argument. If they are all `None`, the disk backend will be chosen.
    Note that It can also register other backend accessor with a given name,
    prefixes, and backend class. In addition, We use the singleton pattern to
    avoid repeated object creation. If the arguments are the same, the same
    object will be returned.

    Args:
        backend (str, optional): The storage backend type. Options are "disk",
            "ceph", "memcached", "lmdb", "http" and "petrel". Default: None.
        prefix (str, optional): The prefix of the registered storage backend.
            Options are "s3", "http", "https". Default: None.

    Examples:
        >>> # only set backend
        >>> file_client = FileClient(backend='petrel')
        >>> # only set prefix
        >>> file_client = FileClient(prefix='s3')
        >>> # set both backend and prefix but use backend to choose client
        >>> file_client = FileClient(backend='petrel', prefix='s3')
        >>> # if the arguments are the same, the same object is returned
        >>> file_client1 = FileClient(backend='petrel')
        >>> file_client1 is file_client
        True

    Attributes:
        client (:obj:`BaseStorageBackend`): The backend object.
    """

    _backends = {
        'disk': HardDiskBackend,
        'hdfs': HDFSBackend,
        'ali': AliBackend
    }
    # This collection is used to record the overridden backends, and when a
    # backend appears in the collection, the singleton pattern is disabled for
    # that backend, because if the singleton pattern is used, then the object
    # returned will be the backend before overwriting
    _overridden_backends = set()
    _overridden_prefixes = set()

    _instances = {}

    def __new__(cls, backend=None, prefix=None, **kwargs):
        if backend is None and prefix is None:
            backend = 'disk'
        
        # concatenate the arguments to a unique key for determining whether
        # objects with the same arguments were created
        arg_key = f'{backend}:{prefix}'
        for key, value in kwargs.items():
            arg_key += f':{key}:{value}'

        # if a backend was overridden, it will create a new object
        if (arg_key in cls._instances
                and backend not in cls._overridden_backends
                and prefix not in cls._overridden_prefixes):
            _instance = cls._instances[arg_key]
        else:
            # create a new object and put it to _instance
            _instance = super().__new__(cls)
            _instance.client = cls._backends[backend](**kwargs)
            cls._instances[arg_key] = _instance

        return _instance

    @property
    def name(self):
        return self.client.name

    @property
    def allow_symlink(self):
        return self.client.allow_symlink

    @classmethod
    def infer_client(cls,
                     file_client_args: Optional[dict] = None,
                     uri: Optional[str] = None) -> 'FileClient':
        """Infer a suitable file client based on the URI and arguments.

        Args:
            file_client_args (dict, optional): Arguments to instantiate a
                FileClient. Default: None.
            uri (str | Path, optional): Uri to be parsed that contains the file
                prefix. Default: None.

        Returns:
            FileClient: Instantiated FileClient object.
        """
        assert file_client_args is not None or uri is not None
        if file_client_args is None:
            if uri.startswith('hdfs'):
                file_client_args = {'backend': 'hdfs'}
            elif uri.startswith('ali'):
                file_client_args = {'backend': 'ali'}
            elif uri.startswith('/') or uri.startswith('./'):
                file_client_args = {'backend': 'disk'}
        if file_client_args is None:
            return None

        return cls(**file_client_args)

    @classmethod
    def _register_backend(cls, name, backend, force=False, prefixes=None):
        if not isinstance(name, str):
            raise TypeError('the backend name should be a string, '
                            f'but got {type(name)}')
        if not inspect.isclass(backend):
            raise TypeError(
                f'backend should be a class but got {type(backend)}')
        if not issubclass(backend, BaseStorageBackend):
            raise TypeError(
                f'backend {backend} is not a subclass of BaseStorageBackend')
        if not force and name in cls._backends:
            raise KeyError(
                f'{name} is already registered as a storage backend, '
                'add "force=True" if you want to override it')

        if name in cls._backends and force:
            cls._overridden_backends.add(name)
        cls._backends[name] = backend

        if prefixes is not None:
            if isinstance(prefixes, str):
                prefixes = [prefixes]
            else:
                assert isinstance(prefixes, (list, tuple))
            for prefix in prefixes:
                if prefix not in cls._prefix_to_backends:
                    cls._prefix_to_backends[prefix] = backend
                elif (prefix in cls._prefix_to_backends) and force:
                    cls._overridden_prefixes.add(prefix)
                    cls._prefix_to_backends[prefix] = backend
                else:
                    raise KeyError(
                        f'{prefix} is already registered as a storage backend,'
                        ' add "force=True" if you want to override it')

    @classmethod
    def register_backend(cls, name, backend=None, force=False, prefixes=None):
        """Register a backend to FileClient.

        This method can be used as a normal class method or a decorator.

        .. code-block:: python

            class NewBackend(BaseStorageBackend):

                def get(self, filepath):
                    return filepath

                def get_text(self, filepath):
                    return filepath

            FileClient.register_backend('new', NewBackend)

        or

        .. code-block:: python

            @FileClient.register_backend('new')
            class NewBackend(BaseStorageBackend):

                def get(self, filepath):
                    return filepath

                def get_text(self, filepath):
                    return filepath

        Args:
            name (str): The name of the registered backend.
            backend (class, optional): The backend class to be registered,
                which must be a subclass of :class:`BaseStorageBackend`.
                When this method is used as a decorator, backend is None.
                Defaults to None.
            force (bool, optional): Whether to override the backend if the name
                has already been registered. Defaults to False.
            prefixes (str or list[str] or tuple[str], optional): The prefixes
                of the registered storage backend. Default: None.
                `New in version 1.3.15.`
        """
        if backend is not None:
            cls._register_backend(
                name, backend, force=force, prefixes=prefixes)
            return

        def _register(backend_cls):
            cls._register_backend(
                name, backend_cls, force=force, prefixes=prefixes)
            return backend_cls

        return _register

    def get(self, filepath: Union[str, Path]) -> Union[bytes, memoryview]:
        """Read data from a given ``filepath`` with 'rb' mode.

        Note:
            There are two types of return values for ``get``, one is ``bytes``
            and the other is ``memoryview``. The advantage of using memoryview
            is that you can avoid copying, and if you want to convert it to
            ``bytes``, you can use ``.tobytes()``.

        Args:
            filepath (str or Path): Path to read data.

        Returns:
            bytes | memoryview: Expected bytes object or a memory view of the
            bytes object.
        """
        return self.client.get(filepath)

    def get_text(self, filepath: Union[str, Path], encoding='utf-8') -> str:
        """Read data from a given ``filepath`` with 'r' mode.

        Args:
            filepath (str or Path): Path to read data.
            encoding (str): The encoding format used to open the ``filepath``.
                Default: 'utf-8'.

        Returns:
            str: Expected text reading from ``filepath``.
        """
        return self.client.get_text(filepath, encoding)

    def put(self, obj: bytes, filepath: Union[str, Path]) -> None:
        """Write data to a given ``filepath`` with 'wb' mode.

        Note:
            ``put`` should create a directory if the directory of ``filepath``
            does not exist.

        Args:
            obj (bytes): Data to be written.
            filepath (str or Path): Path to write data.
        """
        self.client.put(obj, filepath)

    def put_text(self, obj: str, filepath: Union[str, Path]) -> None:
        """Write data to a given ``filepath`` with 'w' mode.

        Note:
            ``put_text`` should create a directory if the directory of
            ``filepath`` does not exist.

        Args:
            obj (str): Data to be written.
            filepath (str or Path): Path to write data.
            encoding (str, optional): The encoding format used to open the
                `filepath`. Default: 'utf-8'.
        """
        self.client.put_text(obj, filepath)

    def remove(self, filepath: Union[str, Path]) -> None:
        """Remove a file.

        Args:
            filepath (str, Path): Path to be removed.
        """
        self.client.remove(filepath)

    def exists(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path exists.

        Args:
            filepath (str or Path): Path to be checked whether exists.

        Returns:
            bool: Return ``True`` if ``filepath`` exists, ``False`` otherwise.
        """
        return self.client.exists(filepath)

    def isdir(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path is a directory.

        Args:
            filepath (str or Path): Path to be checked whether it is a
                directory.

        Returns:
            bool: Return ``True`` if ``filepath`` points to a directory,
            ``False`` otherwise.
        """
        return self.client.isdir(filepath)

    def isfile(self, filepath: Union[str, Path]) -> bool:
        """Check whether a file path is a file.

        Args:
            filepath (str or Path): Path to be checked whether it is a file.

        Returns:
            bool: Return ``True`` if ``filepath`` points to a file, ``False``
            otherwise.
        """
        return self.client.isfile(filepath)

    def join_path(self, filepath: Union[str, Path],
                  *filepaths: Union[str, Path]) -> str:
        """Concatenate all file paths.

        Join one or more filepath components intelligently. The return value
        is the concatenation of filepath and any members of *filepaths.

        Args:
            filepath (str or Path): Path to be concatenated.

        Returns:
            str: The result of concatenation.
        """
        return self.client.join_path(filepath, *filepaths)

    @contextmanager
    def get_local_path(self, filepath: Union[str, Path]) -> Iterable[str]:
        """Download data from ``filepath`` and write the data to local path.

        ``get_local_path`` is decorated by :meth:`contxtlib.contextmanager`. It
        can be called with ``with`` statement, and when exists from the
        ``with`` statement, the temporary path will be released.

        Note:
            If the ``filepath`` is a local path, just return itself.

        .. warning::
            ``get_local_path`` is an experimental interface that may change in
            the future.

        Args:
            filepath (str or Path): Path to be read data.

        Examples:
            >>> file_client = FileClient(prefix='s3')
            >>> with file_client.get_local_path('s3://bucket/abc.jpg') as path:
            ...     # do something here

        Yields:
            Iterable[str]: Only yield one path.
        """
        with self.client.get_local_path(str(filepath)) as local_path:
            yield local_path

    def list_dir_or_file(self,
                         dir_path: Union[str, Path],
                         list_dir: bool = True,
                         list_file: bool = True,
                         suffix: Optional[Union[str, Tuple[str]]] = None,
                         recursive: bool = False) -> Iterator[str]:
        """Scan a directory to find the interested directories or files in
        arbitrary order.

        Note:
            :meth:`list_dir_or_file` returns the path relative to ``dir_path``.

        Args:
            dir_path (str | Path): Path of the directory.
            list_dir (bool): List the directories. Default: True.
            list_file (bool): List the path of files. Default: True.
            suffix (str or tuple[str], optional):  File suffix
                that we are interested in. Default: None.
            recursive (bool): If set to True, recursively scan the
                directory. Default: False.

        Yields:
            Iterable[str]: A relative path to ``dir_path``.
        """
        yield from self.client.list_dir_or_file(dir_path, list_dir, list_file,
                                                suffix, recursive)

    def ls(self, remote_folder):
        return self.client.ls(remote_folder)

    def mkdir(self, remote_path, p=False):
        return self.client.mkdir(remote_path, p)

    def download(self, remote_path, local_path):
        return self.client.download(remote_path, local_path)

    def upload(self, remote_path, local_path, is_exist=False):
        return self.client.upload(remote_path, local_path, is_exist)


def file_client_get(remote_path, local_path):
    status = FileClient.infer_client(uri=remote_path).download(remote_path, local_path)
    return status


def file_client_put(remote_path, local_path, is_exist=False):
    status = FileClient.infer_client(uri=remote_path).upload(remote_path, local_path, is_exist)
    return status


def file_client_ls(remote_folder):
    result = FileClient.infer_client(uri=remote_folder).ls(remote_folder)
    return result


def file_client_mkdir(remote_path, p=False):
    status = FileClient.infer_client(uri=remote_path).mkdir(remote_path, p)
    return status


def file_client_rm(remote_path):
    status = FileClient.infer_client(uri=remote_path).remove(remote_path)
    return status


def file_client_exists(remote_path):
    status = FileClient.infer_client(uri=remote_path).exists(remote_path)
    return status




from __future__ import unicode_literals
from __future__ import division

import numpy as np

def ae(actual, predicted):
    """
    Computes the absolute error.
    This function computes the absolute error between two numbers,
    or for element between a pair of lists or numpy arrays.
    Parameters
    ----------
    actual : int, float, list of numbers, numpy array
             The ground truth value
    predicted : same type as actual
                The predicted value
    Returns
    -------
    score : double or list of doubles
            The absolute error between actual and predicted
    """
    return np.abs(np.array(actual)-np.array(predicted))

def ce(actual, predicted):
    """
    Computes the classification error.
    This function computes the classification error between two lists
    Parameters
    ----------
    actual : list
             A list of the true classes
    predicted : list
                A list of the predicted classes
    Returns
    -------
    score : double
            The classification error between actual and predicted
    """
    return (sum([1.0 for x,y in zip(actual,predicted) if x != y]) /
            len(actual))

def mae(actual, predicted):
    """
    Computes the mean absolute error.
    This function computes the mean absolute error between two lists
    of numbers.
    Parameters
    ----------
    actual : list of numbers, numpy array
             The ground truth value
    predicted : same type as actual
                The predicted value
    Returns
    -------
    score : double
            The mean absolute error between actual and predicted
    """
    return np.mean(ae(actual, predicted))

def mse(actual, predicted):
    """
    Computes the mean squared error.
    This function computes the mean squared error between two lists
    of numbers.
    Parameters
    ----------
    actual : list of numbers, numpy array
             The ground truth value
    predicted : same type as actual
                The predicted value
    Returns
    -------
    score : double
            The mean squared error between actual and predicted
    """
    return np.mean(se(actual, predicted))

def msle(actual, predicted):
    """
    Computes the mean squared log error.
    This function computes the mean squared log error between two lists
    of numbers.
    Parameters
    ----------
    actual : list of numbers, numpy array
             The ground truth value
    predicted : same type as actual
                The predicted value
    Returns
    -------
    score : double
            The mean squared log error between actual and predicted
    """
    return np.mean(sle(actual, predicted))

def rmse(actual, predicted):
    """
    Computes the root mean squared error.
    This function computes the root mean squared error between two lists
    of numbers.
    Parameters
    ----------
    actual : list of numbers, numpy array
             The ground truth value
    predicted : same type as actual
                The predicted value
    Returns
    -------
    score : double
            The root mean squared error between actual and predicted
    """
    return np.sqrt(mse(actual, predicted))

def rmsle(actual, predicted):
    """
    Computes the root mean squared log error.
    This function computes the root mean squared log error between two lists
    of numbers.
    Parameters
    ----------
    actual : list of numbers, numpy array
             The ground truth value
    predicted : same type as actual
                The predicted value
    Returns
    -------
    score : double
            The root mean squared log error between actual and predicted
    """
    return np.sqrt(msle(actual, predicted))

def se(actual, predicted):
    """
    Computes the squared error.
    This function computes the squared error between two numbers,
    or for element between a pair of lists or numpy arrays.
    Parameters
    ----------
    actual : int, float, list of numbers, numpy array
             The ground truth value
    predicted : same type as actual
                The predicted value
    Returns
    -------
    score : double or list of doubles
            The squared error between actual and predicted
    """
    return np.power(np.array(actual)-np.array(predicted), 2)

def sle(actual, predicted):
    """
    Computes the squared log error.
    This function computes the squared log error between two numbers,
    or for element between a pair of lists or numpy arrays.
    Parameters
    ----------
    actual : int, float, list of numbers, numpy array
             The ground truth value
    predicted : same type as actual
                The predicted value
    Returns
    -------
    score : double or list of doubles
            The squared log error between actual and predicted
    """
    return (np.power(np.log(np.array(actual)+1) -
            np.log(np.array(predicted)+1), 2))

def ll(actual, predicted):
    """
    Computes the log likelihood.
    This function computes the log likelihood between two numbers,
    or for element between a pair of lists or numpy arrays.
    Parameters
    ----------
    actual : int, float, list of numbers, numpy array
             The ground truth value
    predicted : same type as actual
                The predicted value
    Returns
    -------
    score : double or list of doubles
            The log likelihood error between actual and predicted
    """
    actual = np.array(actual)
    predicted = np.array(predicted)
    err = np.seterr(all='ignore')
    score = -(actual*np.log(predicted)+(1-actual)*np.log(1-predicted))
    np.seterr(divide=err['divide'], over=err['over'],
              under=err['under'], invalid=err['invalid'])
    if type(score)==np.ndarray:
        score[np.isnan(score)] = 0
    else:
        if np.isnan(score):
            score = 0
    return score

def log_loss(actual, predicted):
    """
    Computes the log loss.
    This function computes the log loss between two lists
    of numbers.
    Parameters
    ----------
    actual : list of numbers, numpy array
             The ground truth value
    predicted : same type as actual
                The predicted value
    Returns
    -------
    score : double
            The log loss between actual and predicted
    """
    return np.mean(ll(actual, predicted))


# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 17:22
# @File    : operator_registry.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function


import types
from collections import namedtuple
from typing import Any, Dict, List

from antgo.pipeline.hparam import param_scope
from antgo.pipeline.engine.uri import URI


def _get_default_namespace():
    with param_scope() as hp:
        return hp().antgo.default_namespace('anon')


class OperatorRegistry:
    """Operator Registry
    """

    REGISTRY: Dict[str, Any] = {}

    def __init__(self) -> None:
        pass

    @staticmethod
    def resolve(name: str) -> Any:
        """
        Resolve operators by name
        """
        for n in [
                name,
                '{}/{}'.format(_get_default_namespace(), name),
                '{}/{}'.format('builtin', name),
        ]:
            if n in OperatorRegistry.REGISTRY:
                return OperatorRegistry.REGISTRY[n]
        return None

    @staticmethod
    def register(
            name: str = None,
            input_schema=None,  # TODO: parse input_schema from code @jie.hou
            output_schema=None,
            flag=None):
        """
        Register a class, function, or callable as a towhee operators.

        Examples:

        1. register a function as operators

        >>> from towhee import register
        >>> @register
        ... def foo(x, y):
        ...     return x+y

        2. register a class as operators

        >>> @register
        ... class foo_cls():
        ...     def __init__(self, x):
        ...         self.x = x
        ...     def __call__(self, y):
        ...         return self.x + y

        By default, function/class name is used as operators name,
        which is used by the operators factory `towhee.ops` to invoke the operators.

        >>> from towhee import ops
        >>> op = ops.foo()
        >>> op(1, 2)
        3

        >>> op = ops.foo_cls(x=2)
        >>> op(3)
        5

        3. register operators with an alternative name:

        >>> @register(name='my_foo')
        ... def foo(x, y):
        ...     return x+y
        >>> ops.my_foo()(1,2)
        3

        Operator URI and Namespace: The URI (unique reference identifier) of an operators has two parts: namespace and name.
        The namespace helps identify one operators and group the operators into various kinds.
        We can specific the namespace when create an operators:

        >>> ops.anon.my_foo()(1,2)
        3

        `anon` is the default namespace to which an operators is registered if no namespace is specified.
        And it's also the default searching namespace for the operators factory.

        You can also specific the fullname, including namespace when register an operators:

        >>> @register(name='my_namespace/my_foo')
        ... def foo(x, y):
        ...     return x+y
        >>> ops.my_namespace.my_foo()(1,2)
        3

        Output Schema:

        >>> @register(name='my_foo', output_schema='value')
        ... def foo(x, y):
        ...     return x+y
        >>> from towhee.hparam import param_scope
        >>> with param_scope('towhee.need_schema=1'):
        ...     ops.my_foo()(1,2)
        Output(value=3)

        Flag: Each operators type, for example: NNOperator and PyOperator, has their own default `flag`:

        >>> from towhee.operators.base import Operator, NNOperator, PyOperator
        >>> from towhee.operators.base import OperatorFlag
        >>> @register
        ... class foo(NNOperator):
        ...     pass
        >>> foo().flag
        <OperatorFlag.REUSEABLE|STATELESS: 6>

        The default flag can be override by `register(flag=someflag)`:

        >>> @register(flag=OperatorFlag.EMPTYFLAG)
        ... class foo(NNOperator):
        ...     pass
        >>> foo().flag
        <OperatorFlag.EMPTYFLAG: 1>

        Args:
            name (str, optional): operators name, will use the class/function name if None.
            input_schema(NamedTuple, optional): input schema for the operators. Defaults to None.
            output_schema(NamedTuple, optional): output schema, will convert the operators output to NamedTuple if not None.
            flag ([OperatorFlag], optional): operators flag. Defaults to OperatorFlag.EMPTYFLAG.

        Returns:
            [type]: [description]
        """
        if callable(name):
            # the decorator is called directly without any arguments,
            # relaunch the register
            cls = name
            return OperatorRegistry.register()(cls)

        if output_schema is None:  # none output schema
            output_schema = namedtuple('Output', 'col0')
        if isinstance(output_schema, str):  # string schema 'col0 col1'
            output_schema = output_schema.split()

        # list schema ['col0', 'col1']
        if isinstance(output_schema, List):
            if len(output_schema) == 0 or isinstance(output_schema[0], str):
                output_schema = namedtuple('Output', output_schema)
        # list schema [(int, (1, )), (np.float32, (-1, -1, 3))] is for triton, do nothing.

        def wrapper(cls):
            metainfo = dict(input_schema=input_schema,
                            output_schema=output_schema,
                            flag=flag)

            nonlocal name
            name = URI(cls.__name__ if name is None else name).resolve_repo(
                _get_default_namespace())

            if isinstance(cls, types.FunctionType):
                OperatorRegistry.REGISTRY[name + '_func'] = cls

            # wrap a callable to a class
            if not isinstance(cls, type) and callable(cls):
                func = cls
                cls = type(
                    cls.__name__, (object, ), {
                        '__call__': lambda _, *arg, **kws: func(*arg, **kws),
                        '__doc__': func.__doc__,
                    })

            if output_schema is not None:
                old_call = cls.__call__

                def wrapper_call(self, *args, **kws):
                    with param_scope() as hp:
                        need_schema = hp().antgo.need_schema(False)
                    if need_schema:
                        return output_schema(old_call(self, *args, **kws))
                    else:
                        return old_call(self, *args, **kws)

                cls.__call__ = wrapper_call
                cls.__abstractmethods__ = set()
            cls.metainfo = metainfo
            if flag is not None:
                cls.flag = property(lambda _: flag)
            OperatorRegistry.REGISTRY[name] = cls

            return cls

        return wrapper


if __name__ == '__main__':
    import doctest

    doctest.testmod(verbose=False)



import sys
import os
import json
import numpy as np
from antgo.utils.sample_gt import *
from antgo.interactcontext import InteractContext
from antgo.framework.helper.fileio import *
from antgo.tools.package import *
import logging
from antgo.framework.helper.task_flag import *


def label_to_studio(src_json_file, tgt_folder, prefix='',  **kwargs):
    # 将标准GT转换到label-studio格式
    if tgt_folder is None:
        tgt_folder = './'
    

    pass


def label_start(src_json_file, tgt_folder, tags, label_type, white_users_str=None, ignore_incomplete=False, root=None, exp=None):
    # 检查参数
    assert(tags is not None)

    if tgt_folder is None:
        tgt_folder = './output'

    # 如果设置root and exp, 则按照规定规则读取等待标注数据。忽略src_json_file。
    if root is not None and exp is not None and root != '' and exp != '':
        # 开始标记
        running_flag(os.path.join(root, 'label', exp))
        if file_client_exists(os.path.join(root, 'label', exp, 'data.tar')):
            if not os.path.exists('./label'):
                os.mkdir('./label')
            
            file_client_get(os.path.join(root, 'label', exp, 'data.tar'), './label/')
            os.system('cd label && tar -xf data.tar')
            # 直接覆盖src_json_file
            src_json_file = './label/annotation.json'
            # 直接覆盖tgt_folder
            tgt_folder = './label/'

    if not os.path.exists(tgt_folder):
        os.makedirs(tgt_folder)

    if not os.path.isdir(tgt_folder):
        logging.error(f'--tgt {tgt_folder} must be a folder.')
        return 

    white_users = {}
    if white_users_str is not None:
        for t in white_users_str.split(','):
            user_name, password = t.split(':')
            white_users.update({
                user_name: {'password': password}
            })
    if len(white_users) == 0:
        white_users = None

    if label_type is None:
        label_type = ''

    assert(label_type in ['', 'RECT','POINT','POLYGON'])

    # label_name:0,label_name;1,...
    label_name_and_label_id_map = {}
    label_id_and_label_name_map = {}
    for label_name_and_label_id in tags.split(','):
        label_name, label_id = label_name_and_label_id.split(':')
        label_name_and_label_id_map[label_name] = int(label_id)
        label_id_and_label_name_map[int(label_id)] = label_name

    inner_category = []
    for label_name, label_id in label_name_and_label_id_map.items():
        inner_category.append({
            'class_name': label_name,
            'class_index': label_id
        })

    # 自动定义服务名字
    src_json_file_name = src_json_file.split('/')[-1].split('.')[0]

    # 启动标注服务，直接保存标准格式
    ctx = InteractContext()
    # 需要在这里设置标注配置信息，如
    # category: 设置类别信息
    # white_users: 设置允许参与的标注人员信息
    # label_type: 设置标注类型，目前仅支持'RECT','POINT','POLYGON'   
    ctx.activelearning.start(src_json_file_name, config={
            'metas':{
                'category': inner_category,
                'level': 'image-level' if label_type == '' else 'instance-level'
            },
            'white_users': white_users,
            'label_type': label_type,   # 设置标注类型，'RECT','POINT','POLYGON'
        }, json_file=src_json_file
    )

    # 切换到标注状态
    ctx.activelearning.labeling()

    # 下载当前标注结果（后台等待标注完成后，再下载）
    result = ctx.activelearning.download()

    # 转换到标准GT格式
    total_gt_list = []
    sgtt = SampleGTTemplate()
    for anno_id, anno_info in enumerate(result):
        standard_gt = sgtt.get()
        standard_gt['image_file'] = '/'.join(anno_info['file_upload'].split('/')[3:])
        standard_gt['tag'] = src_json_file_name

        if len(anno_info['annotations']) == 0:
            continue

        sample_anno_info = anno_info['annotations'][0]
        for sample_anno_instance in sample_anno_info['result']:  
            if sample_anno_instance['type'] == 'RECT':
                standard_gt['height'] = 0 if 'height' not in sample_anno_instance else sample_anno_instance['height']
                standard_gt['width'] = 0 if 'width' not in sample_anno_instance else sample_anno_instance['width']
                            
                bbox_x = sample_anno_instance['value']['x']
                bbox_y = sample_anno_instance['value']['y']
                bbox_width = sample_anno_instance['value']['width']
                bbox_height = sample_anno_instance['value']['height']
                standard_gt['bboxes'].append([bbox_x,bbox_y,bbox_x+bbox_width,bbox_y+bbox_height])

                label_id = sample_anno_instance['value']['labels']
                standard_gt['labels'].append(label_id)
                standard_gt['label_names'].append(label_id_and_label_name_map[label_id])
            elif sample_anno_instance['type'] == 'POLYGON':
                standard_gt['height'] = 0 if 'height' not in sample_anno_instance else sample_anno_instance['height']
                standard_gt['width'] = 0 if 'width' not in sample_anno_instance else sample_anno_instance['width']
                                
                points = []
                for xxyy in sample_anno_instance['value']['points']:
                    points.append([xxyy['x'], xxyy['y']])
                    
                points_array = np.array(points) 
                bbox_x1 = float(np.min(points_array[:,0]))
                bbox_y1 = float(np.min(points_array[:,1]))
                bbox_x2 = float(np.max(points_array[:,0]))
                bbox_y2 = float(np.max(points_array[:,1]))
                standard_gt['bboxes'].append([bbox_x1, bbox_y1, bbox_x2, bbox_y2])

                label_id = sample_anno_instance['value']['labels']
                standard_gt['labels'].append(label_id)
                standard_gt['label_names'].append(label_id_and_label_name_map[label_id])

                standard_gt['segments'].append(points)
                standard_gt['has_segments'].append(1)
            elif sample_anno_instance['type'] == 'POINT':
                pass
            elif sample_anno_instance['type'] == 'CHOICES':
                standard_gt['height'] = 0 if 'height' not in sample_anno_instance else sample_anno_instance['height']
                standard_gt['width'] = 0 if 'width' not in sample_anno_instance else sample_anno_instance['width']
                                
                image_labe_name = ','.join(sample_anno_instance['value']['labels'])
                standard_gt['image_label_name'] = image_labe_name
                standard_gt['image_label'] = label_name_and_label_id_map[image_labe_name] if image_labe_name in label_name_and_label_id_map else -1
 
        if ignore_incomplete:
            if len(standard_gt['bboxes']) == 0 and standard_gt['image_label'] == -1:
                # 忽略空标注样本
                continue
        
        total_gt_list.append(standard_gt)    

    # # debug
    # total_gt_list = []
    # sgtt = SampleGTTemplate()
    # for _ in range(1000):
    #     standard_gt = sgtt.get()
    #     image_files = os.listdir('./label/images')
    #     print('images files number')
    #     print(len(image_files))
    #     standard_gt['image_file'] = f'images/{image_files[0]}'
    #     total_gt_list.append(standard_gt)

    anno_json_file_name = f'{src_json_file_name}_label.json'
    if os.path.exists(os.path.join(tgt_folder,anno_json_file_name)):
        count = 0
        while True:
            anno_json_file_name = f'{src_json_file_name}_label_{count}.json'
            if not os.path.exists(os.path.join(tgt_folder,anno_json_file_name)):
                break
            count += 1

    with open(os.path.join(tgt_folder, anno_json_file_name), 'w',  encoding="utf-8") as fp:
        json.dump(total_gt_list, fp)

    meta_info = sgtt.meta()    
    if os.path.exists(os.path.join(tgt_folder, 'meta.json')):
         with open(os.path.join(tgt_folder, 'meta.json'), 'r', encoding="utf-8") as fp:
             meta_info = json.load(fp)

    meta_info['extent']['total'] = len(total_gt_list)
    with open(os.path.join(tgt_folder, 'meta.json'), 'w', encoding="utf-8") as fp:
        json.dump(meta_info, fp)

    if root is not None and exp is not None and root != '' and exp != '':
        # 仅在设置root and exp后，起作用
        # 1. 打包到tfrecord格式
        package_to_tfrecord(
            os.path.join(tgt_folder, anno_json_file_name), 
            os.path.join(tgt_folder, 'package'),
            prefix=exp,
            size_in_shard=40000,
            thread_num=1
        )

        # 2. 上传到远程
        if not file_client_exists(os.path.join(root, 'dataset', 'label')):
            file_client_mkdir(os.path.join(root, 'dataset', 'label'), True)

        # 打包后的数据
        for file_name in os.listdir(os.path.join(tgt_folder, 'package')):
            file_client_put(
                os.path.join(root, 'dataset', 'label'), 
                os.path.join(tgt_folder, 'package', file_name)
            )

        # 标注文件
        file_client_put(
            os.path.join(root, 'label', exp),
            os.path.join(tgt_folder, anno_json_file_name)
        )

        # 完成标记
        finish_flag(os.path.join(root, 'label', exp))
        
    # 全局结束
    ctx.activelearning.exit()


def label_from_native(src_json_file, tgt_folder, tags, ignore_incomplete=False, **kwargs):
    # 加载原生标注工具的结果
    with open(src_json_file, 'r') as fp:
        result = json.load(fp)
    
    # label_name:0,label_name;1,...
    label_name_and_label_id_map = {}
    label_id_and_label_name_map = {}
    for label_name_and_label_id in tags.split(','):
        label_name, label_id = label_name_and_label_id.split(':')
        label_name_and_label_id_map[label_name] = int(label_id)
        label_id_and_label_name_map[int(label_id)] = label_name
        
    # 转换到标准GT格式
    total_gt_list = []
    src_json_file_name = src_json_file.split('/')[-1].split('.')[0]
    sgtt = SampleGTTemplate()
    for anno_id, anno_info in enumerate(result):
        standard_gt = sgtt.get()
        standard_gt['image_file'] = '/'.join(anno_info['file_upload'].split('/')[3:])
        standard_gt['tag'] = src_json_file_name
        
        if len(anno_info['annotations']) == 0:
            continue
        
        sample_anno_info = anno_info['annotations'][0]
        for sample_anno_instance in sample_anno_info['result']:  
            if sample_anno_instance['type'] == 'RECT':
                standard_gt['height'] = 0 if 'height' not in sample_anno_instance else sample_anno_instance['height']
                standard_gt['width'] = 0 if 'width' not in sample_anno_instance else sample_anno_instance['width']
                            
                bbox_x = sample_anno_instance['value']['x']
                bbox_y = sample_anno_instance['value']['y']
                bbox_width = sample_anno_instance['value']['width']
                bbox_height = sample_anno_instance['value']['height']
                standard_gt['bboxes'].append([bbox_x,bbox_y,bbox_x+bbox_width,bbox_y+bbox_height])

                label_id = sample_anno_instance['value']['labels']
                standard_gt['labels'].append(label_id)
                standard_gt['label_names'].append(label_id_and_label_name_map[label_id])
            elif sample_anno_instance['type'] == 'POLYGON':
                standard_gt['height'] = 0 if 'height' not in sample_anno_instance else sample_anno_instance['height']
                standard_gt['width'] = 0 if 'width' not in sample_anno_instance else sample_anno_instance['width']
                                
                points = []
                for xxyy in sample_anno_instance['value']['points']:
                    points.append([xxyy['x'], xxyy['y']])
                    
                points_array = np.array(points) 
                bbox_x1 = float(np.min(points_array[:,0]))
                bbox_y1 = float(np.min(points_array[:,1]))
                bbox_x2 = float(np.max(points_array[:,0]))
                bbox_y2 = float(np.max(points_array[:,1]))
                standard_gt['bboxes'].append([bbox_x1, bbox_y1, bbox_x2, bbox_y2])

                label_id = sample_anno_instance['value']['labels']
                standard_gt['labels'].append(label_id)
                standard_gt['label_names'].append(label_id_and_label_name_map[label_id])

                standard_gt['segments'].append(points)
                standard_gt['has_segments'].append(1)
            elif sample_anno_instance['type'] == 'POINT':
                pass
            elif sample_anno_instance['type'] == 'CHOICES':
                standard_gt['height'] = 0 if 'height' not in sample_anno_instance else sample_anno_instance['height']
                standard_gt['width'] = 0 if 'width' not in sample_anno_instance else sample_anno_instance['width']
                                
                image_labe_name = ','.join(sample_anno_instance['value']['labels'])
                standard_gt['image_label_name'] = image_labe_name
                standard_gt['image_label'] = label_name_and_label_id_map[image_labe_name] if image_labe_name in label_name_and_label_id_map else -1

        if ignore_incomplete:
            if len(standard_gt['bboxes']) == 0 and standard_gt['image_label'] == -1:
                # 忽略空标注样本
                continue
            
        total_gt_list.append(standard_gt)    
    
    if not os.path.exists(tgt_folder):
        os.makedirs(tgt_folder)
    
    anno_json_file_name = f'{src_json_file_name}_label.json'
    if os.path.exists(os.path.join(tgt_folder,anno_json_file_name)):
        count = 0
        while True:
            anno_json_file_name = f'{src_json_file_name}_label_{count}.json'
            if not os.path.exists(os.path.join(tgt_folder,anno_json_file_name)):
                break
            count += 1
            
    with open(os.path.join(tgt_folder, anno_json_file_name), 'w',  encoding="utf-8") as fp:
        json.dump(total_gt_list, fp)
    
    meta_info = sgtt.meta()    
    if os.path.exists(os.path.join(tgt_folder, 'meta.json')):
         with open(os.path.join(tgt_folder, 'meta.json'), 'r', encoding="utf-8") as fp:
             meta_info = json.load(fp)

    meta_info['extent']['total'] = len(total_gt_list)
    with open(os.path.join(tgt_folder, 'meta.json'), 'w', encoding="utf-8") as fp:
        json.dump(meta_info, fp)    


def label_from_studio(src_json_file, tgt_folder, prefix='', tags=None, ignore_incomplete=False, **kwargs):
    # 将label-studio转换到标注GT
    if tgt_folder is None:
        tgt_folder = './'
    
    if not os.path.exists(tgt_folder):
        os.makedirs(tgt_folder)
        
    if not os.path.isdir(tgt_folder):
        logging.error(f'--tgt {tgt_folder} must be a folder.')
        return 
    
    if tags is None:
        tags = ''
    if prefix is None:
        prefix = ''
    
    # label_name:0,label_name;1,...
    label_name_and_label_id_map = {}
    for label_name_and_label_id in tags.split(','):
        label_name, label_id = label_name_and_label_id.split(':')
        label_name_and_label_id_map[label_name] = int(label_id)
    
    src_json_file_name = src_json_file.split('/')[-1].split('.')[0]
    
    with open(src_json_file, 'r') as fp:
        sample_anno_list = json.load(fp)
    
    total_gt_list = []
    sgtt = SampleGTTemplate()
    for anno_info in sample_anno_list:
        # only use human 0 result
        if len(anno_info['annotations']) == 0:
            continue
        
        group_anno_ids = {}
        original_file = '-'.join(anno_info['file_upload'].split('-')[1:])
        standard_gt = sgtt.get()
        standard_gt['image_file'] = os.path.join(prefix, original_file) if prefix != '' else original_file
        standard_gt['tag'] = src_json_file_name
        sample_anno_info = anno_info['annotations'][0]
        for sample_anno_instance in sample_anno_info['result']:
            if sample_anno_instance['type'] == 'rectanglelabels':
                # 矩形框标注
                standard_gt['height'] = sample_anno_instance['original_height']
                standard_gt['width'] = sample_anno_instance['original_width']
                                
                bbox_x = sample_anno_instance['value']['x'] / 100.0 * standard_gt['width']
                bbox_y = sample_anno_instance['value']['y'] / 100.0 * standard_gt['height']
                bbox_width = sample_anno_instance['value']['width'] / 100.0 * standard_gt['width']
                bbox_height = sample_anno_instance['value']['height'] / 100.0 * standard_gt['height']
                bbox_rotation = sample_anno_instance['value']['rotation']
                standard_gt['bboxes'].append([bbox_x,bbox_y,bbox_x+bbox_width,bbox_y+bbox_height])

                label_name = sample_anno_instance['value']['rectanglelabels'][0]
                standard_gt['labels'].append(label_name_and_label_id_map[label_name] if label_name in label_name_and_label_id_map else -1)
                standard_gt['label_names'].append(label_name)
            elif sample_anno_instance['type'] == 'keypointlabels':
                # 2D关键点标注
                standard_gt['height'] = sample_anno_instance['original_height']
                standard_gt['width'] = sample_anno_instance['original_width']  
                keypoint_x = sample_anno_instance['value']['x'] / 100.0 * standard_gt['width']
                keypoint_y = sample_anno_instance['value']['y'] / 100.0 * standard_gt['height']
                
                # 先把所有关键点保存起来，然后进行组合到单个个体
                label_name = sample_anno_instance['value']['keypointlabels'][0]
                label_id = label_name_and_label_id_map[label_name]
                # label_order = label_name_order[label_name]
                
                sample_anno_id = sample_anno_instance['id']
                sample_parent_anno_id = sample_anno_instance['parentID'] if 'parentID' in sample_anno_instance else ''

                group_anno_ids[sample_anno_id] = {
                    'keypoint_x': keypoint_x,
                    'keypoint_y': keypoint_y,
                    'label_name': label_name,
                    'label_id': label_id,
                    'anno_id': sample_anno_id,
                    'group_anno_id': sample_parent_anno_id
                }
            elif sample_anno_instance['type'] == 'polygonlabels':
                # 分割标注(polygon)
                standard_gt['height'] = sample_anno_instance['original_height']
                standard_gt['width'] = sample_anno_instance['original_width']  
                
                points = sample_anno_instance['value']['points']
                label_name = sample_anno_instance['value']['polygonlabels'][0]
                label_id = label_name_and_label_id_map[label_name]
                
                points_array = np.array(points) 
                points_array[:, 0] = points_array[:, 0] / 100.0 * standard_gt['width']
                points_array[:, 1] = points_array[:, 1] / 100.0 * standard_gt['height']
                points = points_array.tolist()
                
                bbox_x1 = float(np.min(points_array[:,0]))
                bbox_y1 = float(np.min(points_array[:,1]))
                bbox_x2 = float(np.max(points_array[:,0]))
                bbox_y2 = float(np.max(points_array[:,1]))
                standard_gt['bboxes'].append([bbox_x1, bbox_y1, bbox_x2, bbox_y2])

                label_name = sample_anno_instance['value']['polygonlabels'][0]
                standard_gt['labels'].append(label_name_and_label_id_map[label_name] if label_name in label_name_and_label_id_map else -1)
                standard_gt['label_names'].append(label_name)

                standard_gt['segments'].append(points)
                standard_gt['has_segments'].append(1)
                
                assert(len(standard_gt['segments']) == len(standard_gt['bboxes']))
            elif sample_anno_instance['type'] == 'choices':
                # 图片级类别标注
                label_name = ','.join(sample_anno_instance['value']['choices'])
                standard_gt['image_label_name'] = label_name
                standard_gt['image_label'] = label_name_and_label_id_map[label_name] if label_name in label_name_and_label_id_map else -1
       
        if len(group_anno_ids) > 0:
            # 仅对关键点标注启用
            regroup_anno_ids = {}
            for k,v in group_anno_ids.items():
                if v['group_anno_id'] != '': 
                    # 只抽取group id
                    regroup_anno_ids[v['group_anno_id']] = []

            for k,v in group_anno_ids.items():
                if v['group_anno_id'] != '':
                    regroup_anno_ids[v['group_anno_id']].append(v)
                elif v['anno_id'] in regroup_anno_ids:
                    regroup_anno_ids[v['anno_id']].append(v)
            
            # 重新排序每个group
            standard_gt['joints2d'] = [None for _ in range(len(regroup_anno_ids))]
            standard_gt['has_joints2d'] = []
            
            for group_i, group_key in enumerate(regroup_anno_ids.keys()):
                standard_gt['joints2d'][group_i] = [[] for _ in range(len(label_name_and_label_id_map))]
                standard_gt['has_joints2d'].append(1)
                
                for anno_info in regroup_anno_ids[group_key]:
                    label_id = anno_info['label_id']
                    standard_gt['joints2d'][group_i][label_id] = [anno_info['keypoint_x'], anno_info['keypoint_y']]

            # 
            for group_i in range(len(standard_gt['joints2d'])):
                points_array = np.array(standard_gt['joints2d'][group_i]) 
                
                bbox_x1 = float(np.min(points_array[:,0]))
                bbox_y1 = float(np.min(points_array[:,1]))
                bbox_x2 = float(np.max(points_array[:,0]))
                bbox_y2 = float(np.max(points_array[:,1]))
                standard_gt['bboxes'].append([bbox_x1, bbox_y1, bbox_x2, bbox_y2])

                standard_gt['labels'].append(0)
                standard_gt['label_names'].append('unkown')
                
        total_gt_list.append(standard_gt)

    anno_json_file_name = f'{src_json_file_name}_label.json'
    if os.path.exists(os.path.join(tgt_folder,anno_json_file_name)):
        count = 0
        while True:
            anno_json_file_name = f'{src_json_file_name}_label_{count}.json'
            if not os.path.exists(os.path.join(tgt_folder,anno_json_file_name)):
                break
            count += 1
            
    with open(os.path.join(tgt_folder, anno_json_file_name), 'w',  encoding="utf-8") as fp:
        json.dump(total_gt_list, fp)
    
    meta_info = sgtt.meta()    
    if os.path.exists(os.path.join(tgt_folder, 'meta.json')):
         with open(os.path.join(tgt_folder, 'meta.json'), 'r', encoding="utf-8") as fp:
             meta_info = json.load(fp)

    meta_info['extent']['total'] = len(total_gt_list)
    with open(os.path.join(tgt_folder, 'meta.json'), 'w', encoding="utf-8") as fp:
        json.dump(meta_info, fp)  


def label_to_merge(src_json_file, tgt_folder, **kwargs):
    src_json_files = src_json_file.split(',')
    total_list = []
    for json_file in src_json_files:
        if not json_file.endswith('json'):
            logging.warn(f'Ignore {json_file}')
            continue
        
        with open(json_file, 'r', encoding='utf-8') as fp:
            total_list.extend(json.load(fp))

    if not os.path.exists(tgt_folder):
        os.makedirs(tgt_folder)
    
    if not os.path.isdir(tgt_folder):
        logging.error(f'--tgt={tgt_folder} must be a folder')
        return
    
    anno_json_file_name = f'annotation_merge.json'
    if os.path.exists(os.path.join(tgt_folder,anno_json_file_name)):
        count = 0
        while True:
            anno_json_file_name = f'annotation_merge_{count}.json'
            if not os.path.exists(os.path.join(tgt_folder,anno_json_file_name)):
                break
            count += 1
            
    with open(os.path.join(tgt_folder, anno_json_file_name), 'w',  encoding="utf-8") as fp:
        json.dump(total_list, fp)
    
    sgtt = SampleGTTemplate()
    meta_info = sgtt.meta()    
    if os.path.exists(os.path.join(tgt_folder, 'meta.json')):
         with open(os.path.join(tgt_folder, 'meta.json'), 'r', encoding="utf-8") as fp:
             meta_info = json.load(fp)

    meta_info['extent']['total'] = len(total_list)
    with open(os.path.join(tgt_folder, 'meta.json'), 'w', encoding="utf-8") as fp:
        json.dump(meta_info, fp)  

# 测试检测标注
# label_from_studio('/Users/bytedance/Downloads/project-1-at-2023-03-11-02-17-0c19afd8.json', None, '', "Airplane:0,Car:1")
# 测试分割
# label_from_studio('/Users/bytedance/Downloads/project-4-at-2023-03-11-03-41-0557a589.json', None, '', "Airplane:0,Car:1")
# 测试图片标签
# label_from_studio('/Users/bytedance/Downloads/project-2-at-2023-03-11-02-25-96db91e7.json', None, '', "Weapons:0,Violence:1")
# 测试关键点
# label_from_studio('/Users/bytedance/Downloads/project-3-at-2023-03-11-03-01-3a7cf0e3.json', None, '', "Face:0,Nose:1")
# print('a')

# 测试标注
# if __name__ == '__main__':
#     label_start('/Users/bytedance/Downloads/mm/annotation.json', None, 'Car:0', "POLYGON")
#     print('sdf')


# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
import copy

TYPE_ORDER = ["bool", "int32", "int64", "float16", "float32", "float64"]
TYPE_MAPPER = {"fp16": "float16", "fp32": "float32", "fp64": "float64"}

def binary_cross_entropy_with_logits(input,
                                     target,
                                     weight=None,
                                     size_average=None,
                                     reduce=None,
                                     reduction='mean',
                                     pos_weight=None):
    if not reduce or not size_average:
        reduction = "sum"
    input_t = str(input.dtype).lower().strip().split(".")[-1]
    if input_t in TYPE_MAPPER:
        input_t = TYPE_MAPPER[input_t]
    input_index = TYPE_ORDER.index(input_t)
    target_t = str(target.dtype).lower().strip().split(".")[-1]
    if target_t in TYPE_MAPPER:
        target_t = TYPE_MAPPER[target_t]
    target_index = TYPE_ORDER.index(target_t)
    if input_index < target_index:
        real_type = TYPE_ORDER[target_index]
        input = input.cast(real_type)
    else:
        real_type = TYPE_ORDER[input_index]
        target = target.cast(real_type)
    return paddle.nn.functional.binary_cross_entropy_with_logits(
        input, target, weight, reduction, pos_weight)


def avg_pool1d(input,
               kernel_size,
               stride=None,
               padding=0,
               ceil_mode=False,
               count_include_pad=True):
    return paddle.nn.functional.avg_pool1d(
        input,
        kernel_size,
        stride=stride,
        padding=padding,
        ceil_mode=ceil_mode,
        exclusive=not count_include_pad)


def avg_pool2d(input,
               kernel_size,
               stride=None,
               padding=0,
               ceil_mode=False,
               count_include_pad=True,
               divisor_override=None):
    return paddle.nn.functional.avg_pool2d(
        input,
        kernel_size,
        stride=stride,
        padding=padding,
        ceil_mode=ceil_mode,
        exclusive=not count_include_pad,
        divisor_override=divisor_override)


def avg_pool3d(input,
               kernel_size,
               stride=None,
               padding=0,
               ceil_mode=False,
               count_include_pad=True,
               divisor_override=None):
    return paddle.nn.functional.avg_pool3d(
        input,
        kernel_size,
        stride=stride,
        padding=padding,
        ceil_mode=ceil_mode,
        exclusive=not count_include_pad,
        divisor_override=divisor_override)


def dropout(input, p=0.5, training=True, inplace=False):
    return paddle.nn.functional.dropout(input, p=p, training=training)


def interpolate(input,
                size=None,
                scale_factor=None,
                mode='nearest',
                align_corners=None,
                recompute_scale_factor=None):
    return paddle.nn.functional.interpolate(
        input,
        size=size,
        scale_factor=scale_factor,
        mode=mode,
        align_corners=align_corners)


def leaky_relu(input, negative_slope=0.01, inplace=False):
    return paddle.nn.functional.leaky_relu(input, negative_slope=negative_slope)


def log_softmax(input, dim=None, _stacklevel=3, dtype=None):
    return paddle.nn.functional.log_softmax(input, axis=dim, dtype=None)


def mse_loss(input, target, size_average=None, reduce=None, reduction='mean'):
    paddle.nn.functional.mse_loss(input, target, reduction=reduction)


def relu(input, inplace=False):
    return paddle.nn.functional.relu(input)


def smooth_l1_loss(input,
                   target,
                   size_average=None,
                   reduce=None,
                   reduction='mean',
                   beta=1.0):
    paddle.nn.functional.smooth_l1_loss(
        input, target, reduction=reduction, delta=beta)


def softmax(input, dim=None, _stacklevel=3, dtype=None):
    return paddle.nn.functional.softmax(input, axis=dim, dtype=dtype)



import os
from aligo import Aligo

def main():
    ali = Aligo()

    shared_data_info_folder = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'resource', 'dataset')
    # 1.step ade20k
    file_obj = ali.get_file_by_path('/dataset/ade20k/ADEChallengeData2016.zip')
    result = ali.share_file_by_aligo(file_obj)

    with open(os.path.join(shared_data_info_folder, 'ade20k.txt'), 'w') as fp:
        fp.write(result)
    print('finish ade20k')

    # 2.step 3dpw
    file_obj = ali.get_file_by_path('/dataset/3dpw/3dpw.tar')
    result = ali.share_file_by_aligo(file_obj)

    with open(os.path.join(shared_data_info_folder, '3dpw.txt'), 'w') as fp:
        fp.write(result)
    print('finish 3dpw')

    # 3.step cityscapes
    file_obj = ali.get_file_by_path('/dataset/cityscapes/cityscapes.tar')
    result = ali.share_file_by_aligo(file_obj)

    with open(os.path.join(shared_data_info_folder, 'cityscapes.txt'), 'w') as fp:
        fp.write(result)
    print('finish cityscapes')


    # 4.step h36m
    file_obj = ali.get_folder_by_path('/dataset/h36m')
    file_obj = ali.get_file_list(file_obj.file_id)
    result = ali.share_files_by_aligo(file_obj)

    with open(os.path.join(shared_data_info_folder, 'h36m.txt'), 'w') as fp:
        fp.write(result)
    print('finish h36m')

    # 5.step lfw
    file_obj = ali.get_folder_by_path('/dataset/lfw')
    file_obj = ali.get_file_list(file_obj.file_id)
    result = ali.share_files_by_aligo(file_obj)

    with open(os.path.join(shared_data_info_folder,'lfw.txt'), 'w') as fp:
        fp.write(result)
    print('finish lfw')


    # 6.step lip
    file_obj = ali.get_folder_by_path('/dataset/lip')
    file_obj = ali.get_file_list(file_obj.file_id)
    result = ali.share_files_by_aligo(file_obj)

    with open(os.path.join(shared_data_info_folder,'lip.txt'), 'w') as fp:
        fp.write(result)
    print('finish lip')


    # 6.step voc
    file_obj = ali.get_folder_by_path('/dataset/voc')
    file_obj = ali.get_file_list(file_obj.file_id)
    result = ali.share_files_by_aligo(file_obj)

    with open(os.path.join(shared_data_info_folder, 'voc.txt'), 'w') as fp:
        fp.write(result)
    print('finish voc')

    # 7.step coco
    file_obj = ali.get_folder_by_path('/dataset/coco')
    file_obj = ali.get_file_list(file_obj.file_id)
    result = ali.share_files_by_aligo(file_obj)

    with open(os.path.join(shared_data_info_folder,'coco.txt'), 'w') as fp:
        fp.write(result)
    print('finish coco')

    # 7.step flic
    file_obj = ali.get_folder_by_path('/dataset/flic')
    file_obj = ali.get_file_list(file_obj.file_id)
    result = ali.share_files_by_aligo(file_obj)

    with open(os.path.join(shared_data_info_folder, 'flic.txt'), 'w') as fp:
        fp.write(result)
    print('finish flic')

    # 8.step vgg-face2
    file_obj = ali.get_folder_by_path('/dataset/vgg-face2/data')
    file_obj = ali.get_file_list(file_obj.file_id)
    result = ali.share_files_by_aligo(file_obj)

    with open(os.path.join(shared_data_info_folder, 'vgg-face2-data.txt'), 'w') as fp:
        fp.write(result)

    file_obj = ali.get_folder_by_path('/dataset/vgg-face2/meta')
    file_obj = ali.get_file_list(file_obj.file_id)
    result = ali.share_files_by_aligo(file_obj)

    with open(os.path.join(shared_data_info_folder, 'vgg-face2-meta.txt'), 'w') as fp:
        fp.write(result)
    print('finish vgg-face2')

    # 9.step lsp
    file_obj = ali.get_folder_by_path('/dataset/lsp')
    file_obj = ali.get_file_list(file_obj.file_id)
    result = ali.share_files_by_aligo(file_obj)

    with open(os.path.join(shared_data_info_folder, 'lsp.txt'), 'w') as fp:
        fp.write(result)
    print('finish lsp')

    # 10.step mpii
    file_obj = ali.get_folder_by_path('/dataset/mpii')
    file_obj = ali.get_file_list(file_obj.file_id)
    result = ali.share_files_by_aligo(file_obj)

    with open(os.path.join(shared_data_info_folder, 'mpii.txt'), 'w') as fp:
        fp.write(result)
    print('finsh mpii')


if __name__ == '__main__':
  main()



# Copyright (c) OpenMMLab. All rights reserved.
import math

import torch
import torch.nn as nn
from antgo.framework.helper.cnn.bricks.transformer import POSITIONAL_ENCODING
from antgo.framework.helper.base_module import BaseModule


@POSITIONAL_ENCODING.register_module()
class SinePositionalEncoding(BaseModule):
    """Position encoding with sine and cosine functions.

    See `End-to-End Object Detection with Transformers
    <https://arxiv.org/pdf/2005.12872>`_ for details.

    Args:
        num_feats (int): The feature dimension for each position
            along x-axis or y-axis. Note the final returned dimension
            for each position is 2 times of this value.
        temperature (int, optional): The temperature used for scaling
            the position embedding. Defaults to 10000.
        normalize (bool, optional): Whether to normalize the position
            embedding. Defaults to False.
        scale (float, optional): A scale factor that scales the position
            embedding. The scale will be used only when `normalize` is True.
            Defaults to 2*pi.
        eps (float, optional): A value added to the denominator for
            numerical stability. Defaults to 1e-6.
        offset (float): offset add to embed when do the normalization.
            Defaults to 0.
        init_cfg (dict or list[dict], optional): Initialization config dict.
            Default: None
    """

    def __init__(self,
                 num_feats,
                 temperature=10000,
                 normalize=False,
                 scale=2 * math.pi,
                 eps=1e-6,
                 offset=0.,
                 init_cfg=None):
        super(SinePositionalEncoding, self).__init__(init_cfg)
        if normalize:
            assert isinstance(scale, (float, int)), 'when normalize is set,' \
                'scale should be provided and in float or int type, ' \
                f'found {type(scale)}'
        self.num_feats = num_feats
        self.temperature = temperature
        self.normalize = normalize
        self.scale = scale
        self.eps = eps
        self.offset = offset

    def forward(self, mask):
        """Forward function for `SinePositionalEncoding`.

        Args:
            mask (Tensor): ByteTensor mask. Non-zero values representing
                ignored positions, while zero values means valid positions
                for this image. Shape [bs, h, w].

        Returns:
            pos (Tensor): Returned position embedding with shape
                [bs, num_feats*2, h, w].
        """
        # For convenience of exporting to ONNX, it's required to convert
        # `masks` from bool to int.
        mask = mask.to(torch.int)
        not_mask = 1 - mask  # logical_not
        y_embed = not_mask.cumsum(1, dtype=torch.float32)
        x_embed = not_mask.cumsum(2, dtype=torch.float32)
        if self.normalize:
            y_embed = (y_embed + self.offset) / \
                      (y_embed[:, -1:, :] + self.eps) * self.scale
            x_embed = (x_embed + self.offset) / \
                      (x_embed[:, :, -1:] + self.eps) * self.scale
        dim_t = torch.arange(
            self.num_feats, dtype=torch.float32, device=mask.device)
        dim_t = self.temperature**(2 * (dim_t // 2) / self.num_feats)
        pos_x = x_embed[:, :, :, None] / dim_t
        pos_y = y_embed[:, :, :, None] / dim_t
        # use `view` instead of `flatten` for dynamically exporting to ONNX
        B, H, W = mask.size()
        pos_x = torch.stack(
            (pos_x[:, :, :, 0::2].sin(), pos_x[:, :, :, 1::2].cos()),
            dim=4).view(B, H, W, -1)
        pos_y = torch.stack(
            (pos_y[:, :, :, 0::2].sin(), pos_y[:, :, :, 1::2].cos()),
            dim=4).view(B, H, W, -1)
        pos = torch.cat((pos_y, pos_x), dim=3).permute(0, 3, 1, 2)
        return pos

    def __repr__(self):
        """str: a string that describes the module"""
        repr_str = self.__class__.__name__
        repr_str += f'(num_feats={self.num_feats}, '
        repr_str += f'temperature={self.temperature}, '
        repr_str += f'normalize={self.normalize}, '
        repr_str += f'scale={self.scale}, '
        repr_str += f'eps={self.eps})'
        return repr_str


@POSITIONAL_ENCODING.register_module()
class LearnedPositionalEncoding(BaseModule):
    """Position embedding with learnable embedding weights.

    Args:
        num_feats (int): The feature dimension for each position
            along x-axis or y-axis. The final returned dimension for
            each position is 2 times of this value.
        row_num_embed (int, optional): The dictionary size of row embeddings.
            Default 50.
        col_num_embed (int, optional): The dictionary size of col embeddings.
            Default 50.
        init_cfg (dict or list[dict], optional): Initialization config dict.
    """

    def __init__(self,
                 num_feats,
                 row_num_embed=50,
                 col_num_embed=50,
                 init_cfg=dict(type='Uniform', layer='Embedding')):
        super(LearnedPositionalEncoding, self).__init__(init_cfg)
        self.row_embed = nn.Embedding(row_num_embed, num_feats)
        self.col_embed = nn.Embedding(col_num_embed, num_feats)
        self.num_feats = num_feats
        self.row_num_embed = row_num_embed
        self.col_num_embed = col_num_embed

    def forward(self, mask):
        """Forward function for `LearnedPositionalEncoding`.

        Args:
            mask (Tensor): ByteTensor mask. Non-zero values representing
                ignored positions, while zero values means valid positions
                for this image. Shape [bs, h, w].

        Returns:
            pos (Tensor): Returned position embedding with shape
                [bs, num_feats*2, h, w].
        """
        h, w = mask.shape[-2:]
        x = torch.arange(w, device=mask.device)
        y = torch.arange(h, device=mask.device)
        x_embed = self.col_embed(x)
        y_embed = self.row_embed(y)
        pos = torch.cat(
            (x_embed.unsqueeze(0).repeat(h, 1, 1), y_embed.unsqueeze(1).repeat(
                1, w, 1)),
            dim=-1).permute(2, 0,
                            1).unsqueeze(0).repeat(mask.shape[0], 1, 1, 1)
        return pos

    def __repr__(self):
        """str: a string that describes the module"""
        repr_str = self.__class__.__name__
        repr_str += f'(num_feats={self.num_feats}, '
        repr_str += f'row_num_embed={self.row_num_embed}, '
        repr_str += f'col_num_embed={self.col_num_embed})'
        return repr_str



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/11 23:11
# @File    : entity.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from typing import overload


class Entity:
    """
    Entity is the basic data type in DataCollection.

    Users can create an Entity with free schema, which means there is no limit on attribute name and type.
    """

    @overload
    def __init__(self):
        """
        Create an empty entity.
        """
        ...

    def __init__(self, **kwargs):
        """
        Create an Entity with given attributes.
        """
        for k, v in kwargs.items():
            self.__setattr__(k, v)

    def __repr__(self):
        """
        Define the representation of the Entity.

        Examples:

        >>> from antgo.pipeline import Entity
        >>> e = Entity(a = 1, b = 2)
        >>> e
        <Entity dict_keys(['a', 'b'])>
        """
        content = repr(self.__dict__.keys())
        return f'<{self.__class__.__name__} {content.strip()}>'

    def __str__(self) -> str:
        """
        Return the str representation of an Entity.

        Examples:

        >>> from antgo.pipeline import Entity
        >>> e = Entity(a = 1, b = 2)
        >>> str(e)
        "{'a': 1, 'b': 2}"
        """
        return str(self.__dict__)


class EntityView:
    """
    The view to iterate DataFrames.

    Args:
        offset (`int`):
            The offset of an Entity in the table.

    Examples:

    >>> from antgo.pipeline import Entity, DataFrame
    >>> e = [Entity(a=a, b=b) for a,b in zip(range(3), range(3))]
    >>> df = DataFrame(e)
    >>> df = df.to_column()
    >>> df.to_list()[0]
    <EntityView dict_keys(['a', 'b'])>
    >>> df.to_list()[0].a
    0
    >>> df.to_list()[0].b
    0
    """

    def __init__(self, offset: int, table):
        self._offset = offset
        self._table = table

    def __getattr__(self, name):
        value = self._table[name][self._offset]
        try:
            return value.as_py()
        # pylint: disable=bare-except
        except:
            return value

    def __setattr__(self, name, value):
        if name in ('_table', '_offset'):
            self.__dict__[name] = value
            return
        self._table.write(name, self._offset, value)

        if self._offset == len(self._table) - 1:
            self._table.seal()
        self._table = self._table.prepare()

    def __repr__(self):
        """
        Define the representation of the EntityView.

        Examples:

        >>> from antgo.pipeline import Entity, DataFrame
        >>> e = [Entity(a=a, b=b) for a,b in zip(range(5), range(5))]
        >>> df = DataFrame(e)
        >>> df = df.to_column()
        >>> df.to_list()[0]
        <EntityView dict_keys(['a', 'b'])>

        >>> df = DataFrame(e)
        >>> df = df.set_chunksize(2)
        >>> df.to_list()[0]
        <EntityView dict_keys(['a', 'b'])>
        """
        return f'<{self.__class__.__name__} dict_keys({self._table.column_names})>'




# -*- coding: UTF-8 -*-
# @Time    : 2022/5/1 21:31
# @File    : nn.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import torch


class AvgPool1D(torch.nn.AvgPool1d):
  def __init__(self, kernel_size, stride=None, padding=0, exclusive=True, ceil_mode=False, name=None):
    super().__init__(kernel_size=kernel_size,
                     stride=stride,
                     padding=padding,
                     ceil_mode=ceil_mode,
                     count_include_pad=exclusive)


class AvgPool2D(torch.nn.AvgPool2d):
  def __init__(self,
               kernel_size,
               stride=None,
               padding=0,
               ceil_mode=False,
               exclusive=True,
               divisor_override=None,
               data_format='NCHW',
               name=None):
    assert(data_format == 'NCHW')
    super().__init__(
      kernel_size=kernel_size,
      stride=stride,
      padding=padding,
      ceil_mode=ceil_mode,
      count_include_pad=exclusive,
      divisor_override=divisor_override
    )


class AvgPool3D(torch.nn.AvgPool3d):
  def __init__(self,
               kernel_size,
               stride=None,
               padding=0,
               ceil_mode=False,
               exclusive=True,
               divisor_override=None,
               data_format='NCDHW',
               name=None):
    assert(data_format == 'NCDHW')
    super().__init__(
      kernel_size=kernel_size,
      stride=stride,
      padding=padding,
      ceil_mode=ceil_mode,
      count_include_pad=exclusive,
      divisor_override=divisor_override
    )


class BatchNorm(torch.nn.Module):
  def __init__(self,
               num_channels,
               act=None,
               is_test=False,
               momentum=0.9,
               epsilon=1e-05,
               param_attr=None,
               bias_attr=None,
               dtype='float32',
               data_layout='NCHW',
               in_place=False,
               moving_mean_name=None,
               moving_variance_name=None,
               do_model_average_for_mean_and_var=False,
               use_global_stats=False,
               trainable_statistics=False):
    super().__init__()
    self.bn = torch.nn.BatchNorm2d(num_channels,
                                   eps=epsilon,
                                   momentum=1-momentum,
                                   affine=True,
                                   track_running_stats=use_global_stats,
                                   )
    self.act = act
    self.param_attr = param_attr
    self.bias_attr = bias_attr

    if self.act is not None:
      assert(self.act in ['relu', 'relu6'])

    self.init_weight()

  def init_weight(self):
    if self.param_attr is not None and self.param_attr.initializer is not None:
      self.param_attr.initializer(self.bn.weight)

    if self.bias_attr is not None and self.bias_attr.initializer is not None:
      self.bias_attr.initializer(self.bn.bias)

  def forward(self, x):
    x = self.bn(x)
    if self.act is not None:
      if self.act == 'relu':
        x = torch.relu(x)
      elif self.act == 'relu6':
        x = torch.nn.functional.relu6(x)

    return x


class BatchNorm1D(torch.nn.BatchNorm1d):
  def __init__(self,
               num_features,
               momentum=0.9,
               epsilon=1e-05,
               weight_attr=None,
               bias_attr=None,
               data_format='NCL',
               name=None,
               use_global_stats=True):
    momentum = 1-momentum
    super(BatchNorm1D, self).__init__(
      num_features=num_features,
      eps=epsilon,
      momentum=momentum,
      affine=True,
      track_running_stats=use_global_stats,
    )
    self.param_attr = weight_attr
    self.bias_attr = bias_attr
    self.init_weight()

  def init_weight(self):
    if self.param_attr is not None and self.param_attr.initializer is not None:
      self.param_attr.initializer(self.bn.weight)

    if self.bias_attr is not None and self.bias_attr.initializer is not None:
      self.bias_attr.initializer(self.bn.bias)


class BatchNorm2D(torch.nn.BatchNorm2d):
  def __init__(self,
               num_features,
               momentum=0.9,
               epsilon=1e-05,
               weight_attr=None,
               bias_attr=None,
               data_format='NCHW',
               name=None,
               use_global_stats=True):
    momentum = 1 - momentum
    super(BatchNorm2D, self).__init__(
      num_features=num_features,
      eps=epsilon,
      momentum=momentum,
      affine=True,
      track_running_stats=use_global_stats,
    )
    self.param_attr = weight_attr
    self.bias_attr = bias_attr
    self.init_weight()

  def init_weight(self):
    if self.param_attr is not None and self.param_attr.initializer is not None:
      self.param_attr.initializer(self.bn.weight)

    if self.bias_attr is not None and self.bias_attr.initializer is not None:
      self.bias_attr.initializer(self.bn.bias)


class BatchNorm3D(torch.nn.BatchNorm3d):
  def __init__(self,
               num_features,
               momentum=0.9,
               epsilon=1e-05,
               weight_attr=None,
               bias_attr=None,
               data_format='NCDHW',
               name=None,
               use_global_stats=True):
    momentum = 1 - momentum
    super(BatchNorm3D, self).__init__(
      num_features=num_features,
      eps=epsilon,
      momentum=momentum,
      affine=True,
      track_running_stats=use_global_stats,
    )
    self.param_attr = weight_attr
    self.bias_attr = bias_attr
    self.init_weight()

  def init_weight(self):
    if self.param_attr is not None and self.param_attr.initializer is not None:
      self.param_attr.initializer(self.bn.weight)

    if self.bias_attr is not None and self.bias_attr.initializer is not None:
      self.bias_attr.initializer(self.bn.bias)


class BCEWithLogitsLoss(torch.nn.BCEWithLogitsLoss):
  def __init__(self,
               weight=None,
               reduction='mean',
               pos_weight=None,
               name=None):
    super(BCEWithLogitsLoss, self).__init__(
      weight=weight,
      size_average=None,
      reduction=reduction,
      pos_weight=pos_weight
    )


class Pad2D(torch.nn.ConstantPad2d):
  def __init__(self,
               padding,
               mode='constant',
               value=0.0,
               data_format='NCHW',
               name=None):
    assert(mode == 'constant')
    assert(data_format == 'NCHW')
    super(Pad2D, self).__init__(padding=padding, value=value)


class Conv1D(torch.nn.Conv1d):
  def __init__(self,
               in_channels,
               out_channels,
               kernel_size,
               stride=1,
               padding=0,
               dilation=1,
               groups=1,
               padding_mode='zeros',
               weight_attr=None,
               bias_attr=None,
               data_format='NCL'):
    assert (data_format == 'NCL')
    super(Conv1D, self).__init__(in_channels,
                                 out_channels,
                                 kernel_size,
                                 stride=stride,
                                 padding=padding,
                                 dilation=dilation,
                                 groups=groups,
                                 bias=True if bias_attr is not None else False,
                                 padding_mode=padding_mode,
                                 device=None,
                                 dtype=None)
    self.weight_attr = weight_attr
    self.bias_attr = bias_attr
    self.init_weight()

  def init_weight(self):
    if self.weight_attr is not None and self.weight_attr.initializer is not None:
      self.weight_attr.initializer(self.weight)

    if self.bias_attr is not None and self.bias_attr.initializer is not None:
      self.bias_attr.initializer(self.bias)


class Conv2D(torch.nn.Conv2d):
  def __init__(self,
               in_channels,
               out_channels,
               kernel_size,
               stride=1,
               padding=0,
               dilation=1,
               groups=1,
               padding_mode='zeros',
               weight_attr=None,
               bias_attr=None,
               data_format='NCHW'):
    assert(data_format == 'NCHW')
    super(Conv2D, self).__init__(in_channels,
                                 out_channels,
                                 kernel_size,
                                 stride=stride,
                                 padding=padding,
                                 dilation=dilation,
                                 groups=groups,
                                 bias=True if bias_attr is not None else False,
                                 padding_mode=padding_mode,
                                 device=None,
                                 dtype=None)
    self.weight_attr = weight_attr
    self.bias_attr = bias_attr
    self.init_weight()

  def init_weight(self):
    if self.weight_attr is not None and self.weight_attr.initializer is not None:
      self.weight_attr.initializer(self.weight)

    if self.bias_attr is not None and self.bias_attr.initializer is not None:
      self.bias_attr.initializer(self.bias)


class Conv3D(torch.nn.Conv3d):
  def __init__(self,
               in_channels,
               out_channels,
               kernel_size,
               stride=1,
               padding=0,
               dilation=1,
               groups=1,
               padding_mode='zeros',
               weight_attr=None,
               bias_attr=None,
               data_format='NCDHW'):
    assert(data_format == 'NCDHW')
    super(Conv3D, self).__init__(in_channels,
                                 out_channels,
                                 kernel_size,
                                 stride=stride,
                                 padding=padding,
                                 dilation=dilation,
                                 groups=groups,
                                 bias=True if bias_attr is not None else False,
                                 padding_mode=padding_mode,
                                 device=None,
                                 dtype=None)
    self.weight_attr = weight_attr
    self.bias_attr = bias_attr
    self.init_weight()

  def init_weight(self):
    if self.weight_attr is not None and self.weight_attr.initializer is not None:
      self.weight_attr.initializer(self.weight)

    if self.bias_attr is not None and self.bias_attr.initializer is not None:
      self.bias_attr.initializer(self.bias)


class Conv2DTranspose(torch.nn.ConvTranspose2d):
  def __init__(self,
               in_channels,
               out_channels,
               kernel_size,
               stride=1,
               padding=0,
               output_padding=0,
               groups=1,
               dilation=1,
               weight_attr=None,
               bias_attr=None,
               data_format='NCHW'):
    super().__init__(in_channels,
                     out_channels,
                     kernel_size,
                     stride=stride,
                     padding=padding,
                     output_padding=output_padding,
                     groups=groups,
                     bias=True if bias_attr is not None else False,
                     dilation=dilation,
                     padding_mode='zeros',
                     device=None,
                     dtype=None)
    self.weight_attr = weight_attr
    self.bias_attr = bias_attr
    self.init_weight()

  def init_weight(self):
    if self.weight_attr is not None and self.weight_attr.initializer is not None:
      self.weight_attr.initializer(self.weight)

    if self.bias_attr is not None and self.bias_attr.initializer is not None:
      self.bias_attr.initializer(self.bias)


class CrossEntropyLoss(torch.nn.CrossEntropyLoss):
  def __init__(self,
               weight=None,
               ignore_index=- 100,
               reduction='mean',
               soft_label=False,
               axis=- 1,
               name=None):
    super().__init__(
      weight=weight,
      size_average=None,
      ignore_index=ignore_index,
      reduce=None,
      reduction=reduction,
      label_smoothing=0.0)


class Dropout(torch.nn.Dropout):
  def __init__(self,
               p=0.5,
               axis=None,
               mode="upscale_in_train",
               name=None):
    super().__init__(
      p=p,
      inplace=False
    )


class Embedding(torch.nn.Embedding):
  def __init__(self,
               num_embeddings,
               embedding_dim,
               padding_idx=None,
               sparse=False,
               weight_attr=None,
               name=None):
    super().__init__(num_embeddings,
                     embedding_dim,
                     padding_idx=padding_idx,
                     max_norm=None,
                     norm_type=2.0,
                     scale_grad_by_freq=False,
                     sparse=sparse,
                     _weight=None,
                     device=None,
                     dtype=None)
    self.weight_attr = weight_attr
    self.init_weight()

  def init_weight(self):
    if self.weight_attr is not None and self.weight_attr.initializer is not None:
      self.weight_attr.initializer(self.weight)


class GroupNorm(torch.nn.GroupNorm):
  def __init__(self,
               num_groups,
               num_channels,
               epsilon=1e-05,
               weight_attr=None,
               bias_attr=None,
               data_format='NCHW',
               name=None):
    assert(data_format == 'NCHW')
    super().__init__(num_groups,
                     num_channels,
                     eps=epsilon,
                     affine=True,
                     device=None,
                     dtype=None)
    self.param_attr = weight_attr
    self.bias_attr = bias_attr
    self.init_weight()

  def init_weight(self):
    if self.param_attr is not None and self.param_attr.initializer is not None:
      self.param_attr.initializer(self.bn.weight)

    if self.bias_attr is not None and self.bias_attr.initializer is not None:
      self.bias_attr.initializer(self.bn.bias)


class InstanceNorm2D(torch.nn.InstanceNorm2d):
  def __init__(self,
               num_features,
               epsilon=1e-05,
               momentum=0.9,
               weight_attr=None,
               bias_attr=None,
               data_format="NCHW",
               track_running_stats=False,
               name=None):
    super().__init__(num_features,
                     eps=epsilon,
                     momentum=1.0-momentum,
                     affine=True,
                     track_running_stats=track_running_stats,
                     device=None, dtype=None)

    self.param_attr = weight_attr
    self.bias_attr = bias_attr
    self.init_weight()

  def init_weight(self):
    if self.param_attr is not None and self.param_attr.initializer is not None:
      self.param_attr.initializer(self.bn.weight)

    if self.bias_attr is not None and self.bias_attr.initializer is not None:
      self.bias_attr.initializer(self.bn.bias)


class LayerNorm(torch.nn.LayerNorm):
  def __init__(self,
               normalized_shape,
               epsilon=1e-05,
               weight_attr=None,
               bias_attr=None,
               name=None):
    super().__init__(normalized_shape,
                     eps=epsilon,
                     elementwise_affine=True,
                     device=None, dtype=None)

    self.param_attr = weight_attr
    self.bias_attr = bias_attr
    self.init_weight()

  def init_weight(self):
    if self.param_attr is not None and self.param_attr.initializer is not None:
      self.param_attr.initializer(self.bn.weight)

    if self.bias_attr is not None and self.bias_attr.initializer is not None:
      self.bias_attr.initializer(self.bn.bias)


class Linear(torch.nn.Linear):
  def __init__(self,
               in_features,
               out_features,
               weight_attr=None,
               bias_attr=None,
               name=None):
    super().__init__(in_features,
                     out_features,
                     bias=True if bias_attr is not None else False,
                     device=None, dtype=None)
    self.weight_attr = weight_attr
    self.bias_attr = bias_attr
    self.init_weight()

  def init_weight(self):
    if self.weight_attr is not None and self.weight_attr.initializer is not None:
      self.weight_attr.initializer(self.weight)

    if self.bias_attr is not None and self.bias_attr.initializer is not None:
      self.bias_attr.initializer(self.bias)


class L1Loss(torch.nn.L1Loss):
  def __init__(self, reduction='mean', name=None):
    super().__init__(size_average=None,
                     reduce=None,
                     reduction=reduction)


class MaxPool1D(torch.nn.MaxPool1d):
  def __init__(self,
               kernel_size,
               stride=None,
               padding=0,
               return_mask=False,
               ceil_mode=False,
               name=None):
    super().__init__(kernel_size,
                     stride=stride,
                     padding=padding,
                     dilation=1,
                     return_indices=return_mask,
                     ceil_mode=ceil_mode)


class MaxPool2D(torch.nn.MaxPool2d):
  def __init__(self,
               kernel_size,
               stride=None,
               padding=0,
               ceil_mode=False,
               return_mask=False,
               data_format='NCHW',
               name=None):
    super().__init__(kernel_size,
                     stride=stride,
                     padding=padding,
                     dilation=1,
                     return_indices=return_mask,
                     ceil_mode=ceil_mode)


class MaxPool3D(torch.nn.MaxPool3d):
  def __init__(self,
               kernel_size,
               stride=None,
               padding=0,
               ceil_mode=False,
               return_mask=False,
               data_format='NCDHW',
               name=None):
    super().__init__(kernel_size,
                     stride=stride,
                     padding=padding,
                     dilation=1,
                     return_indices=return_mask,
                     ceil_mode=ceil_mode)


class ReLU(torch.nn.ReLU):
  def __init__(self, name=None):
    super().__init__(inplace=False)


class ReLU6(torch.nn.ReLU6):
  def __init__(self, name=None):
    super().__init__(inplace=False)


class Softmax(torch.nn.Softmax):
  def __init__(self, axis=- 1, name=None):
    super().__init__(dim=axis)


class SyncBatchNorm(torch.nn.SyncBatchNorm):
  def __init__(self,
               num_features,
               epsilon=1e-5,
               momentum=0.9,
               weight_attr=None,
               bias_attr=None,
               data_format='NCHW',
               use_global_stats=True,
               name=None):
    super().__init__(num_features,
                     eps=epsilon,
                     momentum=1.0-momentum,
                     affine=True,
                     track_running_stats=use_global_stats,
                     process_group=None,
                     device=None,
                     dtype=None)
    self.param_attr = weight_attr
    self.bias_attr = bias_attr
    self.init_weight()

  def init_weight(self):
    if self.param_attr is not None and self.param_attr.initializer is not None:
      self.param_attr.initializer(self.bn.weight)

    if self.bias_attr is not None and self.bias_attr.initializer is not None:
      self.bias_attr.initializer(self.bn.bias)


class Sequential(torch.nn.Sequential):
  def __init__(self, *args):
    super().__init__(*args)


class Layer(torch.nn.Module):
  def __init__(self):
    super().__init__()


class Functional(object):
  pass

functional = Functional()

def add_functional_function(func):
  setattr(functional, func.__name__, func)

@add_functional_function
def softmax(x, axis=- 1, dtype=None, name=None):
  return torch.softmax(x, dim=axis)

@add_functional_function
def smooth_l1_loss(input, label, reduction='mean', delta=1.0, name=None):
  return torch.nn.functional.smooth_l1_loss(input,
                                            label,
                                            size_average=None,
                                            reduce=None,
                                            reduction=reduction,
                                            beta=delta)

@add_functional_function
def relu(x, name=None):
  return torch.nn.functional.relu(x)

@add_functional_function
def mse_loss(input, label, reduction='mean', name=None):
  return torch.nn.functional.mse_loss(
    input,
    label,
    size_average=None, reduce=None, reduction=reduction
  )

@add_functional_function
def log_softmax(x, axis=- 1, dtype=None, name=None):
  return torch.nn.functional.log_softmax(x, dim=axis)

@add_functional_function
def leaky_relu(x, negative_slope=0.01, name=None):
  return torch.nn.functional.leaky_relu(x, negative_slope=negative_slope)

@add_functional_function
def interpolate(x, size=None, scale_factor=None, mode='nearest', align_corners=False, align_mode=0, data_format='NCHW', name=None):
  assert(data_format == 'NCHW')
  return torch.nn.functional.interpolate(x,
                                         size=size,
                                         scale_factor=scale_factor,
                                         mode=mode,
                                         align_corners=align_corners,
                                         recompute_scale_factor=None)
@add_functional_function
def dropout(x, p=0.5, axis=None, training=True, mode="upscale_in_train", name=None):
  return torch.nn.functional.dropout(x, p=p, training=training)


class Initializer(object):
  pass

initializer = Initializer()

def add_initializer_function(func):
  setattr(initializer, func.__name__, func)

@add_initializer_function
def XavierUniform(fan_in=None, fan_out=None, name=None):
  return lambda x: torch.nn.init.xavier_uniform_(x)

@add_initializer_function
def XavierNormal(fan_in=None, fan_out=None, name=None):
  return lambda x:torch.nn.init.xavier_normal_(x)




# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
from antgo.framework.helper.cnn import ConvModule
from antgo.framework.helper.base_module import BaseModule


class SELayer(BaseModule):
    """Squeeze-and-Excitation Module.

    Args:
        channels (int): The input (and output) channels of the SE layer.
        ratio (int): Squeeze ratio in SELayer, the intermediate channel will be
            ``int(channels/ratio)``. Default: 16.
        conv_cfg (None or dict): Config dict for convolution layer.
            Default: None, which means using conv2d.
        act_cfg (dict or Sequence[dict]): Config dict for activation layer.
            If act_cfg is a dict, two activation layers will be configurated
            by this dict. If act_cfg is a sequence of dicts, the first
            activation layer will be configurated by the first dict and the
            second activation layer will be configurated by the second dict.
            Default: (dict(type='ReLU'), dict(type='Sigmoid'))
        init_cfg (dict or list[dict], optional): Initialization config dict.
            Default: None
    """

    def __init__(self,
                 channels,
                 ratio=16,
                 conv_cfg=None,
                 act_cfg=(dict(type='ReLU'), dict(type='Sigmoid')),
                 init_cfg=None):
        super(SELayer, self).__init__(init_cfg)
        if isinstance(act_cfg, dict):
            act_cfg = (act_cfg, act_cfg)
        assert len(act_cfg) == 2
        # assert mmcv.is_tuple_of(act_cfg, dict)
        self.global_avgpool = nn.AdaptiveAvgPool2d(1)
        self.conv1 = ConvModule(
            in_channels=channels,
            out_channels=int(channels / ratio),
            kernel_size=1,
            stride=1,
            conv_cfg=conv_cfg,
            act_cfg=act_cfg[0])
        self.conv2 = ConvModule(
            in_channels=int(channels / ratio),
            out_channels=channels,
            kernel_size=1,
            stride=1,
            conv_cfg=conv_cfg,
            act_cfg=act_cfg[1])

    def forward(self, x):
        out = self.global_avgpool(x)
        out = self.conv1(out)
        out = self.conv2(out)
        return x * out


class DyReLU(BaseModule):
    """Dynamic ReLU (DyReLU) module.

    See `Dynamic ReLU <https://arxiv.org/abs/2003.10027>`_ for details.
    Current implementation is specialized for task-aware attention in DyHead.
    HSigmoid arguments in default act_cfg follow DyHead official code.
    https://github.com/microsoft/DynamicHead/blob/master/dyhead/dyrelu.py

    Args:
        channels (int): The input (and output) channels of DyReLU module.
        ratio (int): Squeeze ratio in Squeeze-and-Excitation-like module,
            the intermediate channel will be ``int(channels/ratio)``.
            Default: 4.
        conv_cfg (None or dict): Config dict for convolution layer.
            Default: None, which means using conv2d.
        act_cfg (dict or Sequence[dict]): Config dict for activation layer.
            If act_cfg is a dict, two activation layers will be configurated
            by this dict. If act_cfg is a sequence of dicts, the first
            activation layer will be configurated by the first dict and the
            second activation layer will be configurated by the second dict.
            Default: (dict(type='ReLU'), dict(type='HSigmoid', bias=3.0,
            divisor=6.0))
        init_cfg (dict or list[dict], optional): Initialization config dict.
            Default: None
    """

    def __init__(self,
                 channels,
                 ratio=4,
                 conv_cfg=None,
                 act_cfg=(dict(type='ReLU'),
                          dict(type='HSigmoid', bias=3.0, divisor=6.0)),
                 init_cfg=None):
        super().__init__(init_cfg=init_cfg)
        if isinstance(act_cfg, dict):
            act_cfg = (act_cfg, act_cfg)
        assert len(act_cfg) == 2
        # assert mmcv.is_tuple_of(act_cfg, dict)
        self.channels = channels
        self.expansion = 4  # for a1, b1, a2, b2
        self.global_avgpool = nn.AdaptiveAvgPool2d(1)
        self.conv1 = ConvModule(
            in_channels=channels,
            out_channels=int(channels / ratio),
            kernel_size=1,
            stride=1,
            conv_cfg=conv_cfg,
            act_cfg=act_cfg[0])
        self.conv2 = ConvModule(
            in_channels=int(channels / ratio),
            out_channels=channels * self.expansion,
            kernel_size=1,
            stride=1,
            conv_cfg=conv_cfg,
            act_cfg=act_cfg[1])

    def forward(self, x):
        """Forward function."""
        coeffs = self.global_avgpool(x)
        coeffs = self.conv1(coeffs)
        coeffs = self.conv2(coeffs) - 0.5  # value range: [-0.5, 0.5]
        a1, b1, a2, b2 = torch.split(coeffs, self.channels, dim=1)
        a1 = a1 * 2.0 + 1.0  # [-1.0, 1.0] + 1.0
        a2 = a2 * 2.0  # [-1.0, 1.0]
        out = torch.max(x * a1 + b1, x * a2 + b2)
        return out



# Once for All: Train One Network and Specialize it for Efficient Deployment
# Han Cai, Chuang Gan, Tianzhe Wang, Zhekai Zhang, Song Han
# International Conference on Learning Representations (ICLR), 2020.

import os
import numpy as np
import torch
import torch.nn as nn

__all__ = ["AccuracyPredictor"]


class AccuracyPredictor(nn.Module):
    def __init__(
        self,
        arch_encoder,
        hidden_size=400,
        n_layers=3,
        checkpoint_path=None,
        device="cuda:0",
    ):
        super(AccuracyPredictor, self).__init__()
        self.arch_encoder = arch_encoder
        self.hidden_size = hidden_size
        self.n_layers = n_layers
        self.device = device

        # build layers
        layers = []
        for i in range(self.n_layers):
            layers.append(
                nn.Sequential(
                    nn.Linear(
                        self.arch_encoder.n_dim if i == 0 else self.hidden_size,
                        self.hidden_size,
                    ),
                    nn.ReLU(inplace=True),
                )
            )
        layers.append(nn.Linear(self.hidden_size, 1, bias=False))
        self.layers = nn.Sequential(*layers)
        self.base_acc = nn.Parameter(
            torch.zeros(1, device=self.device), requires_grad=False
        )

        if checkpoint_path is not None and os.path.exists(checkpoint_path):
            checkpoint = torch.load(checkpoint_path, map_location="cpu")
            if "state_dict" in checkpoint:
                checkpoint = checkpoint["state_dict"]
            self.load_state_dict(checkpoint)
            print("Loaded checkpoint from %s" % checkpoint_path)

        self.layers = self.layers.to(self.device)

    def forward(self, x):
        y = self.layers(x).squeeze()
        return y + self.base_acc

    def predict_acc(self, arch_dict_list):
        X = [self.arch_encoder.arch2feature(arch_dict) for arch_dict in arch_dict_list]
        X = torch.tensor(np.array(X)).float().to(self.device)
        return self.forward(X)



# Copyright (c) OpenMMLab. All rights reserved.
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import functools
import os
import socket
import subprocess
from collections import OrderedDict

import torch
import torch.multiprocessing as mp
from torch import distributed as dist
from torch._utils import (_flatten_dense_tensors, _take_tensors,
                          _unflatten_dense_tensors)


def _find_free_port():
    # Copied from https://github.com/facebookresearch/detectron2/blob/main/detectron2/engine/launch.py # noqa: E501
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # Binding to port 0 will cause the OS to find an available port for us
    sock.bind(('', 0))
    port = sock.getsockname()[1]
    sock.close()
    # NOTE: there is still a chance the port could be taken by other processes.
    return port


def _is_free_port(port):
    ips = socket.gethostbyname_ex(socket.gethostname())[-1]
    ips.append('localhost')
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return all(s.connect_ex((ip, port)) != 0 for ip in ips)


def init_dist(backend='nccl', **kwargs):
    if mp.get_start_method(allow_none=True) is None:
        mp.set_start_method('spawn')
    _init_dist_pytorch(backend, **kwargs)


def _init_dist_pytorch(backend, **kwargs):
    # TODO: use local_rank instead of rank % num_gpus
    rank = int(os.environ['RANK'])
    num_gpus = torch.cuda.device_count()
    torch.cuda.set_device(rank % num_gpus)
    dist.init_process_group(backend=backend, **kwargs)


def get_dist_info():
    if dist.is_available() and dist.is_initialized():
        rank = dist.get_rank()
        world_size = dist.get_world_size()
    else:
        rank = 0
        world_size = 1
    return rank, world_size


def master_only(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        rank, _ = get_dist_info()
        if rank == 0:
            return func(*args, **kwargs)

    return wrapper


def allreduce_params(params, coalesce=True, bucket_size_mb=-1):
    """Allreduce parameters.

    Args:
        params (list[torch.Parameters]): List of parameters or buffers of a
            model.
        coalesce (bool, optional): Whether allreduce parameters as a whole.
            Defaults to True.
        bucket_size_mb (int, optional): Size of bucket, the unit is MB.
            Defaults to -1.
    """
    _, world_size = get_dist_info()
    if world_size == 1:
        return
    params = [param.data for param in params]
    if coalesce:
        _allreduce_coalesced(params, world_size, bucket_size_mb)
    else:
        for tensor in params:
            dist.all_reduce(tensor.div_(world_size))


def allreduce_grads(params, coalesce=True, bucket_size_mb=-1):
    """Allreduce gradients.

    Args:
        params (list[torch.Parameters]): List of parameters of a model
        coalesce (bool, optional): Whether allreduce parameters as a whole.
            Defaults to True.
        bucket_size_mb (int, optional): Size of bucket, the unit is MB.
            Defaults to -1.
    """
    grads = [
        param.grad.data for param in params
        if param.requires_grad and param.grad is not None
    ]
    _, world_size = get_dist_info()
    if world_size == 1:
        return
    if coalesce:
        _allreduce_coalesced(grads, world_size, bucket_size_mb)
    else:
        for tensor in grads:
            dist.all_reduce(tensor.div_(world_size))


def _allreduce_coalesced(tensors, world_size, bucket_size_mb=-1):
    if bucket_size_mb > 0:
        bucket_size_bytes = bucket_size_mb * 1024 * 1024
        buckets = _take_tensors(tensors, bucket_size_bytes)
    else:
        buckets = OrderedDict()
        for tensor in tensors:
            tp = tensor.type()
            if tp not in buckets:
                buckets[tp] = []
            buckets[tp].append(tensor)
        buckets = buckets.values()

    for bucket in buckets:
        flat_tensors = _flatten_dense_tensors(bucket)
        dist.all_reduce(flat_tensors)
        flat_tensors.div_(world_size)
        for tensor, synced in zip(
                bucket, _unflatten_dense_tensors(flat_tensors, bucket)):
            tensor.copy_(synced)



# -*- coding: UTF-8 -*-
# Time: 1/6/18
# File: yesno_crowdsource.py
# Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.measures.crowdsource import *
from bs4 import BeautifulSoup


default = {'AntYesNoCrowdsource': ('YesorNo', '')}


class AntYesNoCrowdsource(AntCrowdsource):
  def __init__(self, task):
    super(AntYesNoCrowdsource, self).__init__(task, 'YesorNo')

    # DATA_TAG
    # 控制如何从recoder中拾取数据显示在页面上
    self.left_tag = getattr(self.task, 'yes_or_no_left', 'PREDICT')
    self.right_tag = getattr(self.task, 'yes_or_no_right', 'GT')

    # set _client_query_html
    self.client_html_template = 'yesno_crowdsource.html'
    # {HTML_ELEM: DATA_TAG, HTML_ELEM: DATA_TAG, ...}
    self.client_keywords_template = {'A': self.left_tag, 'B': self.right_tag}

    # set _client_query_data
    self.client_query_data = {'QUERY': {self.left_tag: 'IMAGE',
                                        self.right_tag: 'IMAGE'}}
    self._is_support_rank = True
    self._crowdsource_title = 'YesorNo'
    self._crowdsource_type = "YN"
  
  def __where_in_table(self, worksite, element_id):
    if type(element_id) != list:
      element_id = [element_id]

    output = []
    soup = BeautifulSoup(worksite)
    for element in element_id:
      element_node = soup.find(id=element)
      container_id = element_node.parent.get('id')
      output.append(int(container_id))

    return output

  def ground_truth_response(self, client_id, query_index, record_db):
    # CONCLUSION, WORKSITE
    user_worksite = self._client_response_record[client_id]['RESPONSE'][query_index]['WORKSITE']

    # 'PREDICT','GT' position in table
    xy = self.__where_in_table(user_worksite, [self.left_tag, self.right_tag])

    user_gt_response = ''
    if int(xy[0]) < int(xy[1]):
      user_gt_response = 'RIGHT is REAL'
    else:
      user_gt_response = 'LEFT is REAL'

    # parse user_worksite (resource)
    return {'CUSTOM_GT': {'DATA': user_gt_response, 'TYPE': 'TEXT'}}

  def eva(self, data=None, label=None):
    if len(self._client_response_record) == 0:
      with open(os.path.join(self.dump_dir, 'crowdsource_record.txt')) as fp:
        content = fp.read()
        self._client_response_record = json.loads(content)

    score = None
    score_count = None
    person_count = 0
    for _, client_response in self._client_response_record.items():
      response_data = client_response['RESPONSE']
      query_index_map = client_response['ID']

      if score is None:
        score = np.zeros((len(query_index_map), len(self._client_response_record)))
        score_count = np.zeros((len(query_index_map), len(self._client_response_record)))
      
      for query_index, client_data in enumerate(response_data):
        if client_data is None:
          # some clients may have incomplete response
          continue

        real_query_index = query_index_map[query_index]
        client_worksite = client_data['WORKSITE']
        client_conclusion = client_data['CONCLUSION']
        
        xy = self.__where_in_table(client_worksite, [self.left_tag, self.right_tag])
        score_count[real_query_index, person_count] = 1.0
        if int(xy[0]) < int(xy[1]):
          if client_conclusion == 'RIGHT':
            score[real_query_index, person_count] = 1.0
          else:
            score[real_query_index, person_count] = 0.0
        else:
          if client_conclusion == 'LEFT':
            score[real_query_index, person_count] = 1.0
          else:
            score[real_query_index, person_count] = 0.0

      person_count += 1

    denominator_s = np.sum(score_count, axis=1)
    numerator_s = np.sum(score, axis=1)
    evaluation_score = np.mean(numerator_s / denominator_s)
    return {'statistic': {'name': self.name,
                          'value': [{'name': self.name, 'value': float(evaluation_score), 'type': 'SCALAR', 'score_map': score_count}]}, }


from .formatting import IAnyToTensor, IImageToTensor
from .transforms import INormalize, IRandomErasing, IRandomFlip, IRandomResizedCrop
from .builder import *


__all__ = [
    'IAnyToTensor', 'IImageToTensor', 'INormalize',
    'IRandomErasing', 'IRandomFlip', 'IRandomResizedCrop'
]


# 优化器配置
optimizer = dict(type='SGD', lr=0.05,  weight_decay=5e-4, momentum=0.9, nesterov=True)
optimizer_config = dict(grad_clip=None)

# 学习率调度配置
lr_config = dict(
    policy='CosineAnnealing',
    min_lr=1e-5,
)

# 日志配置
log_config = dict(
    interval=5,    
    hooks=[
        dict(type='TextLoggerHook'),
    ])

# 自定义HOOKS
custom_hooks = [
    dict(
        type='EMAHook'
    )
]

# 模型配置
model = dict(
    type='KeypointNet',
    backbone=dict(
        type='KetNetF',
        architecture='resnet34',
        in_channels=3
    ),
    head=dict(
        type='SimpleQNetLite0PoseHeatmap2D',
        num_deconv_filters=[128, 128, 128],
        num_joints=14
    )
)

# 描述模型基本信息
info = dict(
    backbone=dict(
        channels=[96,128,160],
        shapes=[16,8,4]
    )
)

# checkpoint配置
checkpoint_config = dict(interval=1, out_dir='./output/')       

# 数据配置
data=dict(
    train=dict(
        type='LSP',
        train_or_test='train',
        dir='./lsp_dataset',
        pipeline=[
            dict(type="ConvertRandomObjJointsAndOffset", input_size=(128,128), heatmap_size=(16,16), num_joints=14),
            dict(type='ToTensor', keys=['image']),
            dict(type='Normalize', mean=(0.491400, 0.482158, 0.4465231), std=(0.247032, 0.243485, 0.2615877), keys=['image']),
        ],
        inputs_def=dict(
            fields=['image', 'heatmap', 'offset_x', 'offset_y', 'heatmap_weight', 'joints_vis', 'joints2d', 'bboxes'],
        )
    ),
    train_dataloader=dict(
        samples_per_gpu=32, 
        workers_per_gpu=2,
        drop_last=True,
        shuffle=True,
    ),
    val=dict(
        type='LSP',
        train_or_test='val',
        dir='./lsp_dataset',
        pipeline=[
            dict(type="ConvertRandomObjJointsAndOffset", input_size=(128,128), heatmap_size=(16,16), num_joints=14, with_random=False),
            dict(type='ToTensor', keys=['image']),
            dict(type='Normalize', mean=(0.491400, 0.482158, 0.4465231), std=(0.247032, 0.243485, 0.2615877), keys=['image']),
        ],
        inputs_def=dict(
            fields=['image', 'heatmap', 'offset_x', 'offset_y', 'heatmap_weight', 'joints_vis', 'joints2d', 'bboxes'],
        )
    ),
    val_dataloader=dict(
        samples_per_gpu=128, 
        workers_per_gpu=1,
        drop_last=False,
        shuffle=False,
    ),
    test=dict(
        type='LSP',
        train_or_test='test',
        dir='./lsp_dataset',
        pipeline=[
            dict(type="ConvertRandomObjJointsAndOffset", input_size=(128,128), heatmap_size=(16,16), num_joints=14, with_random=False),
            dict(type='ToTensor', keys=['image']),
            dict(type='Normalize', mean=(0.491400, 0.482158, 0.4465231), std=(0.247032, 0.243485, 0.2615877), keys=['image']),
        ],
        inputs_def=dict(
            fields=['image', 'heatmap', 'offset_x', 'offset_y', 'heatmap_weight', 'joints_vis', 'joints2d', 'bboxes'],
        )
    ),
    test_dataloader=dict(
        samples_per_gpu=128, 
        workers_per_gpu=1,
        drop_last=False,
        shuffle=False,
    )
)

# 评估方案配置
evaluation=dict(out_dir='./output/', interval=1, metric=dict(type='OKS'), save_best='oks', rule='greater')

# 导出配置
export=dict(
    input_shape_list = [[1,3,128,128]],
    input_name_list=["image"],
    output_name_list=["heatmap", "offset"]
)

max_epochs = 600


"""Register for Custom Operator."""
from .. import glue


def register(*args, **kwargs):
    """The Wrapper to regiser Custom Operator

    Parameters
    ----------
    *args
    **kwargs
    """
    return glue.register(*args, **kwargs)



from __future__ import unicode_literals
from __future__ import division
from __future__ import print_function
from contextlib import contextmanager
from antgo.dataflow.dataset import *
import antvis.client.mlogger as mlogger
import os


class Params(object):
  def __init__(self, params={}):
    assert(type(params) == dict)
    # if params is not None:
    #   for k, v in params.items():
    #     if k != 'self':
    #       setattr(self, k, v)
    # self.__dict__.update(params)
    self._params = params

  def define(self, k, v=None):
    # setattr(self, k, v)
    self._params[k] = v

  def get(self, item=None, default=None):
    if item is None:
      return self._params

    if item not in self._params:
      return default

    return self._params[item]

  def __getattr__(self, item):
    if item not in self._params:
      return None
    if type(self._params[item]) == dict:
      return Params(self._params[item])

    return self._params[item]

  def keys(self):
    return self._params.keys()

  def items(self):
    return self._params.items()


global_context = None


def get_global_context():
  global global_context
  assert(global_context is not None)
  return global_context


class _Block(object):
  def __init__(self, name):
    self._name = name
    self._activate = True
    
  @property
  def activate(self):
    return self._activate
  @activate.setter
  def activate(self, val):
    self._activate = val
  
  @property
  def name(self):
    return self._name
  
  def __enter__(self):
    return self
  
  def __exit__(self, exc_type, exc_val, exc_tb):
    pass


class _DataProcessor(object):
  def __init__(self):
    self.processor_sequence = []

  def add(self, *args):
    self.processor_sequence.extend(args)

  def iterator(self, source):
    upper = source
    for processor_cls in self.processor_sequence:
      upper = processor_cls(Node.inputs(upper))

    return upper.iterator_value()


class Context(object):
  def __init__(self, interact_mode=False):
    global global_context
    self.training_process_callback = None
    self.infer_process_callback = None
    self._data_processor = _DataProcessor()
    self.running_recorder = None
    self.context_params = None
    
    self.pid = str(os.getpid())
    
    global_context = self
    self.context_ant = None
    self._stage = ""

    self.trainer_callbacks = []
    self.clear_callbacks = []

    self.data_source = None
    self._blocks = []
    self._blocks_status = {}
    self._stoppable_threads = []
    
    self._from_experiment = None
    self._model = None

    self._devices = []

    self._quiet = False
    self._debug = False

    self._name = ''
    self._task_factory = None
    self._data_factory = None
    self._main_folder = None
    self._experiment_uuid = None
    self._is_interact_mode = interact_mode

    # 注册用户数据
    self.register_obj = {}

  def is_interact_mode(self):
    return self._is_interact_mode

  def wait_until_clear(self):
    for stoppable_thread in self._stoppable_threads:
      stoppable_thread.stop()
      if stoppable_thread.stop_condition is not None:
        with stoppable_thread.stop_condition:
          stoppable_thread.stop_condition.notifyAll()
      stoppable_thread.join()
    self._stoppable_threads = []

    for clear_func in self.clear_callbacks:
      clear_func()
    
    global global_context
    global_context = None
    # clear
    self.training_process_callback = None
    self.infer_process_callback = None

    self._data_processor = None

    self.running_recorder = None
    self.context_params = None
    self.context_ant = None
    self._stage = ""
    self.trainer_callbacks = []
    self._data_source = None

  @property
  def name(self):
    return self._name

  @name.setter
  def name(self, val):
    self._name = val

  @property
  def ant(self):
    return self.context_ant

  @ant.setter
  def ant(self, val):
    self.context_ant = val

  @property
  def stage(self):
    return self._stage

  @stage.setter
  def stage(self, val):
    # reset experiment stage
    self._stage = val

  @property
  def devices(self):
    return self._devices

  @devices.setter
  def devices(self, val):
    self._devices = val

  @property
  def model(self):
    return self._model
  
  @model.setter
  def model(self, val):
    self._model = val

  def register(self, **kwargs):
    self.register_obj.update(kwargs)

  def register_at(self, key):
    return self.register_obj.get(key, None)

  @property
  def training_process(self):
    return self.training_process_callback

  @training_process.setter
  def training_process(self, callback):
    self.training_process_callback = callback

  def call_training_process(self, data_source, dump_dir):
    is_inner_set = False
    if self.recorder is not None and self.recorder.dump_dir == None:
      if dump_dir != '':
        if not os.path.exists(os.path.join(dump_dir, 'record')):
          os.makedirs(os.path.join(dump_dir, 'record'))

      self.recorder.dump_dir = os.path.join(dump_dir, 'record')
      is_inner_set = True
    
    self.data_source = data_source
    self.training_process(data_source, dump_dir)

    if self.recorder is not None:
      if self.recorder.dump_dir is not None and is_inner_set:
        self.recorder.dump_dir = None

  @property
  def infer_process(self):
    return self.infer_process_callback

  @infer_process.setter
  def infer_process(self, callback):
    self.infer_process_callback = callback

  def call_infer_process(self, data_source, dump_dir):
    is_inner_set = False
    if self.recorder is not None and self.recorder.dump_dir is None:
      if dump_dir != '':
        if not os.path.exists(os.path.join(dump_dir, 'record')):
          os.makedirs(os.path.join(dump_dir, 'record'))

        self.recorder.dump_dir = os.path.join(dump_dir, 'record')
        is_inner_set = True
      
    self.data_source = data_source
    self.infer_process(data_source, dump_dir)

    if self.recorder is not None:
      if self.recorder.dump_dir is not None and is_inner_set:
        self.recorder.dump_dir = None

  @property
  def data_processor(self):
    return self._data_processor

  @property
  def recorder(self):
    return self.running_recorder

  @recorder.setter
  def recorder(self, callback):
    self.running_recorder = callback

  @property
  def params(self):
    return self.context_params

  @params.setter
  def params(self, val):
    self.context_params = Params(val)

  def registry_trainer_callback(self, key, value, condition, func):
    # condition: equal, less, greater or mod
    self.trainer_callbacks.append((key, value, condition, func))

  @property
  def registried_trainer_callbacks(self):
    return self.trainer_callbacks

  def registry_clear_callback(self, func):
    self.clear_callbacks.append(func)
  
  @property
  def data_source(self):
    return self._data_source
  
  @data_source.setter
  def data_source(self, val):
    self._data_source = val
  
  def register_stoppable_thread(self, stoppable_thread):
    self._stoppable_threads.append(stoppable_thread)
  
  def block(self, name):
    for b in self._blocks:
      if name == b.name:
        return b
  
    model_block = _Block(name)
    self._blocks.append(model_block)
    model_block.activate = True
    if name in self._blocks_status and not self._blocks_status[name]:
      model_block.activate = False
    return model_block
  
  def blocks(self):
    return self._blocks
    
  def activate_block(self, name):
    self._blocks_status[name] = True
    
  def deactivate_block(self, name):
    self._blocks_status[name] = False
  
  @property
  def from_experiment(self):
    return self._from_experiment

  @from_experiment.setter
  def from_experiment(self, experiment):
    self._from_experiment = experiment

  @property
  def quiet(self):
    return self._quiet

  @quiet.setter
  def quiet(self, val):
    self._quiet = val

  @property
  def debug(self):
    return self._debug

  @debug.setter
  def debug(self, val):
    self._debug = val

  @property
  def data_factory(self):
    return self._data_factory

  @data_factory.setter
  def data_factory(self, val):
    self._data_factory = val

  @property
  def task_factory(self):
    return self._task_factory

  @task_factory.setter
  def task_factory(self, val):
    self._task_factory = val

  @property
  def main_folder(self):
    return self._main_folder

  @main_folder.setter
  def main_folder(self, val):
    self._main_folder = val

  @property
  def experiment_uuid(self):
    return self._experiment_uuid

  @experiment_uuid.setter
  def experiment_uuid(self, val):
    self._experiment_uuid = val


from functools import partial

import torch

TORCH_VERSION = torch.__version__


def is_cuda_available() -> bool:
    return torch.cuda.is_available()


IS_CUDA_AVAILABLE = is_cuda_available()


def is_rocm_pytorch() -> bool:
    is_rocm = False
    if TORCH_VERSION != 'parrots':
        try:
            from torch.utils.cpp_extension import ROCM_HOME
            is_rocm = True if ((torch.version.hip is not None) and
                               (ROCM_HOME is not None)) else False
        except ImportError:
            pass
    return is_rocm


def _get_cuda_home():
    if TORCH_VERSION == 'parrots':
        from parrots.utils.build_extension import CUDA_HOME
    else:
        if is_rocm_pytorch():
            from torch.utils.cpp_extension import ROCM_HOME
            CUDA_HOME = ROCM_HOME
        else:
            from torch.utils.cpp_extension import CUDA_HOME
    return CUDA_HOME


def get_build_config():
    if TORCH_VERSION == 'parrots':
        from parrots.config import get_build_info
        return get_build_info()
    else:
        return torch.__config__.show()


def _get_conv():
    if TORCH_VERSION == 'parrots':
        from parrots.nn.modules.conv import _ConvNd, _ConvTransposeMixin
    else:
        from torch.nn.modules.conv import _ConvNd, _ConvTransposeMixin
    return _ConvNd, _ConvTransposeMixin


def _get_dataloader():
    if TORCH_VERSION == 'parrots':
        from torch.utils.data import DataLoader, PoolDataLoader
    else:
        from torch.utils.data import DataLoader
        PoolDataLoader = DataLoader
    return DataLoader, PoolDataLoader


def _get_extension():
    if TORCH_VERSION == 'parrots':
        from parrots.utils.build_extension import BuildExtension, Extension
        CppExtension = partial(Extension, cuda=False)
        CUDAExtension = partial(Extension, cuda=True)
    else:
        from torch.utils.cpp_extension import (BuildExtension, CppExtension,
                                               CUDAExtension)
    return BuildExtension, CppExtension, CUDAExtension


def _get_pool():
    if TORCH_VERSION == 'parrots':
        from parrots.nn.modules.pool import (_AdaptiveAvgPoolNd,
                                             _AdaptiveMaxPoolNd, _AvgPoolNd,
                                             _MaxPoolNd)
    else:
        from torch.nn.modules.pooling import (_AdaptiveAvgPoolNd,
                                              _AdaptiveMaxPoolNd, _AvgPoolNd,
                                              _MaxPoolNd)
    return _AdaptiveAvgPoolNd, _AdaptiveMaxPoolNd, _AvgPoolNd, _MaxPoolNd


def _get_norm():
    if TORCH_VERSION == 'parrots':
        from parrots.nn.modules.batchnorm import _BatchNorm, _InstanceNorm
        SyncBatchNorm_ = torch.nn.SyncBatchNorm2d
    else:
        from torch.nn.modules.batchnorm import _BatchNorm
        from torch.nn.modules.instancenorm import _InstanceNorm
        SyncBatchNorm_ = torch.nn.SyncBatchNorm
    return _BatchNorm, _InstanceNorm, SyncBatchNorm_


_ConvNd, _ConvTransposeMixin = _get_conv()
DataLoader, PoolDataLoader = _get_dataloader()
BuildExtension, CppExtension, CUDAExtension = _get_extension()
_BatchNorm, _InstanceNorm, SyncBatchNorm_ = _get_norm()
_AdaptiveAvgPoolNd, _AdaptiveMaxPoolNd, _AvgPoolNd, _MaxPoolNd = _get_pool()


class SyncBatchNorm(SyncBatchNorm_):

    def _check_input_dim(self, input):
        if TORCH_VERSION == 'parrots':
            if input.dim() < 2:
                raise ValueError(
                    f'expected at least 2D input (got {input.dim()}D input)')
        else:
            super()._check_input_dim(input)



import os
import json
import logging


def prepare_extra_config(task_name, project_info):
    # 根据当前的任务种类（supervised, semi-supervised, distillation, activelearning）
    # 自动使用项目信息补充任务的额外配置，（如数据）
    # 复合任务标记
    extra_config = {}
    if task_name == "supervised":
        # 检查是否有标签数据  
        if len(project_info['dataset']['train']['label']) == 0:
            logging.error(f"No eough label data.")
            return None
                
        # 扩展数据源 (有监督训练，仅涉及标签数据和伪标签数据)
        # label, pseudo-label
        extra_config['source'] = {
            "label": project_info["dataset"]["train"]["label"],
            "pseudo-label": project_info["dataset"]["train"]["pseudo-label"],
        }
    elif task_name == "activelearning":
        # 检查是否有主动学习配置
        if len(project_info['tool']['activelearning']['config']) == 0:
            logging.error(f"Missing {task_name} config, couldnt launch task")
            return None

        # 检查是否有无标签数据
        if len(project_info['dataset']['train']['unlabel']) == 0:
            logging.error(f"No eough unlabel data.")
            return None
        
        # 扩展数据源 (主动学习仅涉及无标签数据)
        # unlabel
        extra_config['source'] = {
            "unlabel": project_info["dataset"]["train"]["unlabel"]
        }

        # 扩展模型配置/优化器/学习率等
        extra_config.update(project_info['tool']['activelearning']['config'])  
    elif task_name == "semi-supervised":
        # 检查是否有半监督学习配置
        if len(project_info['tool']['semi']['config']) == 0:
            logging.error(f"Missing {task_name} config, couldnt launch task")
            return None

        # 检查是否有无标签数据
        if len(project_info['dataset']['train']['unlabel']) == 0:
            logging.error(f"No eough unlabel data.")
            return None
        
        # 扩展数据源（半监督学习涉及，标签数据，伪标签数据和无标签数据）
        # label, pseudo-label, unlabel
        extra_config['source'] = {
            "label": project_info["dataset"]["train"]["label"],
            "pseudo-label": project_info["dataset"]["train"]["pseudo-label"] if "pseudo-label" in project_info["dataset"]["train"] else [],
            "unlabel": project_info["dataset"]["train"]["unlabel"] if "unlabel" in project_info["dataset"]["train"] else []
        }

        # 扩展模型配置/优化器/学习率等
        extra_config.update( project_info['tool']['semi']['config'])
    elif task_name == "distillation":
        # 检查是否有蒸馏学习配置
        if len(project_info['tool']['distillation']['config']) == 0:
            logging.error(f"Missing {task_name} config, couldnt launch task")
            return None

        # 检查是否有标签数据  
        if len(project_info['dataset']['train']['label']) == 0:
            logging.error(f"No enough label data.")
            return None

        # 扩展数据源 (蒸馏学习涉及，标签数据，伪标签数据)
        # label, pseudo-label, unlabel
        extra_config['source'] = {
            "label": project_info["dataset"]["train"]["label"],
            "pseudo-label": project_info["dataset"]["train"]["pseudo-label"],
        }

        # 扩展模型配置/优化器/学习率等
        extra_config.update(project_info['tool']['distillation']['config'])
        
        # 更新专家模型配置
        if len(project_info['expert']) == 0:
            logging.error("No enough expert model config")
            return None
        # 专家模型配置样式
        # {
        #     'config': dict(model=dict(), info=dict()),
        #     'checkpoint': '',
        # }
        extra_config['model']['model']['teacher'] = project_info['expert'][0]['config']['model']
        extra_config['model']['init_cfg'] = {'checkpoint': project_info['expert'][0]['checkpoint']}
        
        info =  project_info['expert'][0]['config']['info']
        if 'channels' in extra_config['model']['train_cfg']['teacher']:
            align_name = extra_config['model']['train_cfg']['align']
            extra_config['model']['train_cfg']['teacher']['channels'] = info.get(align_name).get('channels')
        if 'shapes' in extra_config['model']['train_cfg']['teacher']:
            align_name = extra_config['model']['train_cfg']['align']
            extra_config['model']['train_cfg']['teacher']['shapes'] = info.get(align_name).get('shapes')

    return extra_config



# Copyright (c) OpenMMLab. All rights reserved.
"""Helpers for random number generators."""
import numpy as np


def ensure_rng(rng=None):
    """Coerces input into a random number generator.

    If the input is None, then a global random state is returned.

    If the input is a numeric value, then that is used as a seed to construct a
    random state. Otherwise the input is returned as-is.

    Adapted from [1]_.

    Args:
        rng (int | numpy.random.RandomState | None):
            if None, then defaults to the global rng. Otherwise this can be an
            integer or a RandomState class
    Returns:
        (numpy.random.RandomState) : rng -
            a numpy random number generator

    References:
        .. [1] https://gitlab.kitware.com/computer-vision/kwarray/blob/master/kwarray/util_random.py#L270  # noqa: E501
    """

    if rng is None:
        rng = np.random.mtrand._rand
    elif isinstance(rng, int):
        rng = np.random.RandomState(rng)
    else:
        rng = rng
    return rng



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta, abstractmethod


class BaseAssigner(metaclass=ABCMeta):
    """Base assigner that assigns boxes to ground truth boxes."""

    @abstractmethod
    def assign(self, bboxes, gt_bboxes, gt_bboxes_ignore=None, gt_labels=None):
        """Assign boxes to either a ground truth boxes or a negative boxes."""



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/24 20:26
# @File    : segmentor.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import numpy as np
import mmcv
from mmseg.apis import inference_segmentor, init_segmentor
import os
import subprocess
import sys


class Segmentor(object):
  def __init__(self, config_file, checkpoint_file=None, device='cpu'):
    if not config_file.endswith('.py'):
      config_file = f'{config_file}.py'
    config_name = config_file.split('.')[0]
    model_folder = os.path.join(os.environ['HOME'], '.antgo', 'models', 'mmsegmentation')
    if not os.path.exists(model_folder):
      os.makedirs(model_folder)

    model_file = os.path.join(model_folder, config_file)
    if not os.path.exists(model_file):
      subprocess.check_call([sys.executable, '-m', 'mim', 'download', 'mmsegmentation', '--config', config_name, '--dest', model_folder])

    if checkpoint_file is None:
      for f in os.listdir(model_folder):
        if f.startswith(config_name) and f.endswith('.pth'):
          checkpoint_file = f
          break
    assert(checkpoint_file is not None)

    config_file = os.path.join(model_folder, config_file)
    checkpoint_file = os.path.join(model_folder, checkpoint_file)

    self.model = init_segmentor(config_file, checkpoint_file, device=device)

  def __call__(self, *args, **kwargs):
    result = inference_segmentor(self.model, args[0])

    # label
    return result[0]



# -*- coding: UTF-8 -*-
# @Time    : 2020-06-10 12:18
# @File    : base_server.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import tornado.httpserver
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado import web, gen
from tornado import httpclient
import tornado.web
import json
from tornado.escape import json_decode
import time


# error code
class RESPONSE_STATUS_CODE:
    SUCCESS = 200                   # [GET]
    RESOURCE_SUCCESS_CREATED = 201  # 新建、修改数据成功 [POST/PUT/PATCH]
    RESOURCE_SUCCESS_DELETE = 204   # 数据删除成功
    RESOURCE_NOT_FOUND = 404        # 未获得指定数据
    RESOURCE_GONE = 410             # 请求的资源已经被永久删除，不可得到 [GET]
    EXECUTE_ACCEPTED = 202          # 任务等待执行
    EXECUTE_UNAUTHORIZED = 401      # 未授权执行
    EXECUTE_FORBIDDEN = 403         # 拥有授权，但访问禁止
    REQUEST_NOT_ACCEPTED = 406      # 用户请求格式错误 [GET]
    REQUEST_INVALID = 400           # 用户请求数据存在问题 [POST/PUT/PATCH]
    INTERNAL_SERVER_ERROR = 500     # 服务器发生错误


class BaseHandler(tornado.web.RequestHandler):
  """Base Handler class with access to common methods and properties."""

  def set_default_headers(self):
    self.set_header("Access-Control-Allow-Origin", "http://localhost:8080")
    self.set_header('Access-Control-Allow-Headers', '*')
    self.set_header("Access-Control-Allow-Methods", "POST,GET,PUT,DELETE,PATCH,OPTIONS")
    self.set_header("Access-Control-Expose-Headers", "Content-Disposition")
    self.set_header("Access-Control-Allow-Credentials", "true")

  @property
  def name(self):
    return self.settings.get('name', '-')

  @property
  def html_template(self):
    return self.settings['html_template']

  @property
  def keywords_template(self):
    return self.settings.get('keywords_template', {})

  @property
  def data_folder(self):
    return self.settings['data_folder']

  @property
  def db(self):
    return self.settings['db']

  @property
  def token(self):
    return self.settings['token']

  @property
  def task_name(self):
    return self.settings['task_name']

  @property
  def task_type(self):
    return self.settings['task_type']

  @property
  def upload_folder(self):
    return self.settings.get('upload', None)

  @property
  def download_folder(self):
    return self.settings.get('download', None)

  @property
  def dump_folder(self):
    return self.settings.get('dump_path', None)

  @property
  def request_queue(self):
    return self.settings.get('request_queue', None)

  @property
  def response_queue(self):
    return self.settings.get('response_queue', None)

  def get_json_argument(self, name, default=None):
    args = json_decode(self.request.body)
    # name = to_unicode(name)
    if name in args:
      return args[name]
    elif default is not None:
      return default
    else:
      raise tornado.web.MissingArgumentError(name)

  def timestamp_2_str(self, now_time):
    return time.strftime("%Y-%m-%dx%H-%M-%S", time.localtime(now_time))

  def response(self, status_code=RESPONSE_STATUS_CODE.SUCCESS, message='', content={}, status=None):
    self.set_status(status_code)

    if status is None:
      status = 'OK'
      if status_code not in [200, 201, 204]:
        status = 'ERROR'

    self.write(json.dumps({
      'status': status,
      'message': message,
      'content': content
    }))

  def set_login_cookie(self, user):
    """Set login cookies for the Hub and single-user server."""

    # create and set a new cookie token for the hub
    # if not self.get_current_user_cookie():
    self._set_user_cookie(user, 'antgo')

  def _set_user_cookie(self, user, server):
    # tornado <4.2 have a bug that consider secure==True as soon as
    # 'secure' kwarg is passed to set_secure_cookie
    if self.request.protocol == 'https':
      kwargs = {'secure': True}
    else:
      kwargs = {}
    # if self.subdomain_host:
    #     kwargs['domain'] = self.domain

    # self.log.debug("Setting cookie for %s: %s, %s", user.name, server, json.dumps(kwargs))
    self.set_secure_cookie(
      server,
      user['cookie_id'],
      **kwargs
    )

  def get_current_user_cookie(self):
    """get_current_user from a cookie token"""
    return self._user_for_cookie("antgo")

  def get_current_user(self):
    """get current username"""
    # 判断是否使用白盒用户验证机制
    if self.db.get('white_users', None) is None:
      if 'default_user' not in self.settings:
        default_user = {
          'full_name': 'ANTGO',
          'short_name': 'A',
          'labeling_sample': -1,
          'start_time': -1
        }
        self.settings['default_user'] = default_user

      return self.settings['default_user']
    user = self.get_current_user_cookie()
    return user

  def _user_for_cookie(self, cookie_name, cookie_value=None):
    """Get the User for a given cookie, if there is one"""
    # cookie_id = self.get_secure_cookie(
    #     cookie_name,
    #     cookie_value,
    #     max_age_days=self.cookie_max_age_days,
    # )
    cookie_id = self.get_secure_cookie(
      cookie_name,
      cookie_value
    )

    def clear():
      self.clear_cookie(cookie_name, path='/')

    if cookie_id is None:
      if self.get_cookie(cookie_name):
        clear()
      return

    cookie_id = cookie_id.decode('utf8', 'replace')
    # u = self.db.query(orm.User).filter(orm.User.cookie_id == cookie_id).first()
    if cookie_id not in self.db['users']:
      return None

    return self.db['users'][cookie_id]
